from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import four_scalars, compare_scalars
from .maximum import maximum_value
from .minimum import minimum_value

def range_value(data):
    """
    Determines the range of a data set (i.e., the difference between its largest value and its smallest value)

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    interval : float
        Range of data set

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.maximum.maximum_value`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Range of set: :math:`R = a_{max} - a_{min}`
    
        - :math:`a_{min} \\leq a_j, \\forall a_j \\in a_i`
        - :math:`a_{max} \\geq a_j, \\forall a_j \\in a_i`

    - |range|

    Examples
    --------
    Import `range_value` function from `regressions` library
        >>> from regressions.statistics.ranges import range_value
    Determine the range of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> range_even = range_value([21, 53, 3, 68, 43, 9, 72, 19, 20, 1])
        >>> print(range_even)
        71.0
    Determine the range of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> range_odd = range_value([12, 81, 13, 8, 42, 72, 91, 20, 20])
        >>> print(range_odd)
        83.0
    """
    # Handle input errors
    vector_of_scalars(data)

    # Determine maximum and minimum of input
    max_value = maximum_value(data)
    min_value = minimum_value(data)

    # Calculate difference between maximum and minimum
    difference = max_value - min_value

    # Convert difference to float
    result = float(difference)
    return result

def shift_into_range(initial_value, periodic_unit, minimum, maximum):
    """
    Adjusts an intial value to one within a particular range by increasing or decreasing its value by a specified unit

    Parameters
    ----------
    initial_value : int or float
        Starting value to adjust to fit into a range
    periodic_unit : int or float
        Unit by which the initial value should be incrementally increased or decreased to fit into a range
    minimum : int or float
        Lower bound of range into which the initial value must be adjusted (final value should be greater than or equal to minimum)
    maximum : int or float
        Upper bound of range into which the initial value must be adjusted (final value should be less than or equal to maximum)

    Raises
    ------
    TypeError
        Arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument

    Returns
    -------
    final_value : float
        Value within range that only differs from the initial value by a an integral multiple of the periodic unit

    See Also
    --------
    :func:`~regressions.analyses.points.shifted_points_within_range`, :func:`~regressions.analyses.mean_values.mean_values_derivative`, :func:`~regressions.analyses.mean_values.mean_values_integral`

    Notes
    -----
    - Initial value: :math:`v_i`
    - Periodic unit: :math:`\\lambda`
    - Lower bound of range: :math:`b_l`
    - Upper bound of range: :math:`b_u`
    - Set of all values derived from initial value and periodic unit: :math:`g = \\{ v \\mid v = v_i + \\lambda\\cdot{k} \\}`

        - :math:`k \\in \\mathbb{Z}`
    
    - Final value: :math:`v_f \\geq b_l \\cap v_f \\leq b_u \\cap v_f \\in g`

    Examples
    --------
    Import `shift_into_range` function from `regressions` library
        >>> from regressions.statistics.ranges import shift_into_range
    Adjust the number 7 to a value between 20 and 30, based on a periodic unit of 8
        >>> final_value_int = shift_into_range(7, 8, 20, 30)
        >>> print(final_value_int)
        23.0
    Adjust the number 524.62 to a value between 138.29 and 213.86, based on a periodic unit of 23.91
        >>> final_value_float = shift_into_range(524.62, 23.91, 138.29, 213.86)
        >>> print(final_value_float)
        213.78999999999974
    """
    # Handle input errors
    four_scalars(initial_value, periodic_unit, minimum, maximum)
    compare_scalars(minimum, maximum, 'third', 'fourth')

    # Set input value to alternative value
    alternative_initial_value = initial_value
    
    # Handle positive periodic units
    if periodic_unit > 0:
        # Decrease value till below maximum
        while alternative_initial_value > maximum:
            alternative_initial_value -= periodic_unit

        # Increase value till above minimum
        while alternative_initial_value < minimum:
            alternative_initial_value += periodic_unit
    
    # Handle negative periodic units
    else:
        # Decrease value till below maximum
        while alternative_initial_value > maximum:
            alternative_initial_value += periodic_unit

        # Increase value till above minimum
        while alternative_initial_value < minimum:
            alternative_initial_value -= periodic_unit

    # Convert final value to float
    final_value = float(alternative_initial_value)
    return final_value


from regressions.errors.matrices import square_matrix
from regressions.errors.scalars import whole_number

def inner_determinant(matrix, row, column):
    # Handle input errors
    square_matrix(matrix)
    whole_number(row, 'second')
    whole_number(column, 'third')

    # Create intermediary dictionary and list to return
    storage = {}
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix)):
        if m != row:
            storage[m] = []

            # Iterate over inner lists of input
            for n in range(len(matrix[0])):
                if n!= column:
                    storage[m].append(matrix[m][n])
    
    # Iterate over keys in dictionary
    for key in storage:
        result.append(storage[key])
    
    # Return result
    return result

def linear_determinant(matrix, result = 0):
    """
    Calculate the determinant of a matrix

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        First argument must contain the same amount of lists as the amount of elements contained within its first list
    
    Returns
    -------
    determinant : float
        Determinant of a matrix

    Warning
    -------
    Function has factorial time complexity; not recommended for matrices larger than 5-by-5

    See Also
    --------
    :func:`~regressions.matrices.minors.matrix_of_minors`, :func:`~regressions.matrices.inverse.inverse_matrix`

    Notes
    -----
    - Original matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,j} & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,j} & a_{2,n} \\\\ a_{i,1} & a_{i,2} & \\cdots & a_{i,j} & a_{i,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,j} & a_{m,n} \\end{bmatrix}`
    - Determinant of matrix (if :math:`\\mathbf{A}` contains an odd number of columns): :math:`|\\mathbf{A}| = a_{1,1}\\cdot{|\\mathbf{A}_{1,1}|} - a_{1,2}\\cdot{|\\mathbf{A}_{1,2}|} + \\cdots - a_{1,j}\\cdot{|\\mathbf{A}_{1,j}|} + a_{1,n}\\cdot{|\\mathbf{A}_{1,n}|}`
    - Determinant of matrix (if :math:`\\mathbf{A}` contains an even number of columns): :math:`|\\mathbf{A}| = a_{1,1}\\cdot{|\\mathbf{A}_{1,1}|} - a_{1,2}\\cdot{|\\mathbf{A}_{1,2}|} + \\cdots + a_{1,j}\\cdot{|\\mathbf{A}_{1,j}|} - a_{1,n}\\cdot{|\\mathbf{A}_{1,n}|}`
    - |determinant|

    Examples
    --------
    Import `linear_determinant` function from `regressions` library
        >>> from regressions.matrices.determinant import linear_determinant
    Calculate the determinant of [[1, 2], [3, 4]]
        >>> determinant_2x2 = linear_determinant([[1, 2], [3, 4]])
        >>> print(determinant_2x2)
        -2.0
    Calculate the determinant of [[2, 3, 5], [7, 11, 13], [17, 19, 23]]
        >>> determinant_3x3 = linear_determinant([[2, 3, 5], [7, 11, 13], [17, 19, 23]])
        >>> print(determinant_3x3)
        -78.0
    """
    # Handle input errors
    square_matrix(matrix)
    whole_number(result, 'second')

    # Handle base case
    if len(matrix) == 1:
        result += float(matrix[0][0])
        return result
    
    # Handle recursive case
    else:
        # Create intermediary variables
        alternating = []
        minors = []
        leads = matrix[0]

        # Iterate over first inner list of input
        for i in range(len(leads)):
            minors.append(inner_determinant(matrix, 0, i))
            if i % 2 == 0:
                alternating.append(leads[i])
            else:
                alternating.append(-1 * leads[i])
        
        # Iterate over alternating
        for j in range(len(alternating)):
            result += alternating[j] * linear_determinant(minors[j])
    
    # Convert result to float
    floated_result = float(result)
    return floated_result


from numpy import exp, inf
from scipy.optimize import curve_fit
from regressions.errors.matrices import matrix_of_scalars
from regressions.errors.vectors import long_vector
from regressions.errors.scalars import positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.vectors.dimension import single_dimension
from regressions.analyses.equations.logistic import logistic_equation
from regressions.analyses.derivatives.logistic import logistic_derivatives
from regressions.analyses.integrals.logistic import logistic_integral
from regressions.analyses.points import key_coordinates
from regressions.analyses.accumulation import accumulated_area
from regressions.analyses.mean_values import average_values
from regressions.statistics.halve import half_dimension
from regressions.statistics.mean import mean_value
from regressions.statistics.summary import five_number_summary
from regressions.statistics.correlation import correlation_coefficient
from regressions.statistics.rounding import rounded_value

def logistic_model(data, precision = 4):
    """
    Generates a logistic regression model from a given data set

    Parameters
    ----------
    data : list of lists of int or float
        List of lists of numbers representing a collection of coordinate pairs; it must include at least 10 pairs
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        First argument must contain at least 10 elements
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    model['constants'] : list of float
        Coefficients of the resultant logistic model; the first element is the carrying capacity, the second element is the growth rate, and the third element is the sigmoid's midpoint
    model['evaluations']['equation'] : func
        Function that evaluates the equation of the logistic model at a given numeric input (e.g., model['evaluations']['equation'](10) would evaluate the equation of the logistic model when the independent variable is 10)
    model['evaluations']['derivative'] : func
        Function that evaluates the first derivative of the logistic model at a given numeric input (e.g., model['evaluations']['derivative'](10) would evaluate the first derivative of the logistic model when the independent variable is 10)
    model['evaluations']['integral'] : func
        Function that evaluates the integral of the logistic model at a given numeric input (e.g., model['evaluations']['integral'](10) would evaluate the integral of the logistic model when the independent variable is 10)
    model['points']['roots'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the x-intercepts of the logistic model (will always be `None`)
    model['points']['maxima'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the maxima of the logistic model (will always be `None`)
    model['points']['minima'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the minima of the logistic model (will always be `None`)
    model['points']['inflections'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the inflection points of the logistic model (will contain exactly one point)
    model['accumulations']['range'] : float
        Total area under the curve represented by the logistic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided (i.e., over the range)
    model['accumulations']['iqr'] : float
        Total area under the curve represented by the logistic model between the first and third quartiles of all the independent coordinates originally provided (i.e., over the interquartile range)
    model['averages']['range']['average_value_derivative'] : float
        Average rate of change of the curve represented by the logistic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided
    model['averages']['range']['mean_values_derivative'] : list of float
        All points between the smallest independent coordinate originally provided and the largest independent coordinate originally provided where their instantaneous rate of change equals the function's average rate of change over that interval
    model['averages']['range']['average_value_integral'] : float
        Average value of the curve represented by the logistic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided
    model['averages']['range']['mean_values_integral'] : list of float
        All points between the smallest independent coordinate originally provided and the largest independent coordinate originally provided where their value equals the function's average value over that interval
    model['averages']['iqr']['average_value_derivative'] : float
        Average rate of change of the curve represented by the logistic model between the first and third quartiles of all the independent coordinates originally provided
    model['averages']['iqr']['mean_values_derivative'] : list of float
        All points between the first and third quartiles of all the independent coordinates originally provided where their instantaneous rate of change equals the function's average rate of change over that interval
    model['averages']['iqr']['average_value_integral'] : float
        Average value of the curve represented by the logistic model between the first and third quartiles of all the independent coordinates originally provided
    model['averages']['iqr']['mean_values_integral'] : list of float
        All points between the first and third quartiles of all the independent coordinates originally provided where their value equals the function's average value over that interval
    model['correlation'] : float
        Correlation coefficient indicating how well the model fits the original data set (values range between 0.0, implying no fit, and 1.0, implying a perfect fit)

    See Also
    --------
    :func:`~regressions.analyses.equations.logistic.logistic_equation`, :func:`~regressions.analyses.derivatives.logistic.logistic_derivatives`, :func:`~regressions.analyses.integrals.logistic.logistic_integral`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.statistics.correlation.correlation_coefficient`, :func:`~regressions.execute.run_all`

    Notes
    -----
    - Provided ordered pairs for the data set: :math:`p_i = \\{ (p_{1,x}, p_{1,y}), (p_{2,x}, p_{2,y}), \\cdots, (p_{n,x}, p_{n,y}) \\}`
    - Provided values for the independent variable: :math:`X_i = \\{ p_{1,x}, p_{2,x}, \\cdots, p_{n,x} \\}`
    - Provided values for the dependent variable: :math:`Y_i = \\{ p_{1,y}, p_{2,y}, \\cdots, p_{n,y} \\}`
    - Minimum value of the provided values for the independent variable: :math:`X_{min} \\leq p_{j,x}, \\forall p_{j,x} \\in X_i`
    - Maximum value of the provided values for the independent variable: :math:`X_{max} \\geq p_{j,x}, \\forall p_{j,x} \\in X_i`
    - First quartile of the provided values for the independent variable: :math:`X_{Q1}`
    - Third quartile of the provided values for the independent variable: :math:`X_{Q3}`
    - Mean of all provided values for the dependent variable: :math:`\\bar{y} = \\frac{1}{n}\\cdot{\\sum\\limits_{i=1}^n Y_i}`
    - Resultant values for the coefficients of the logistic model: :math:`C_i = \\{ a, b, c \\}`
    - Standard form for the equation of the logistic model: :math:`f(x) = \\frac{a}{1 + \\text{e}^{-b\\cdot(x - c)}}`
    - First derivative of the logistic model: :math:`f'(x) = \\frac{ab\\cdot{\\text{e}^{-b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^2}`
    - Second derivative of the logistic model: :math:`f''(x) = \\frac{2ab^2\\cdot{\\text{e}^{-2b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^3} - \\frac{ab^2\\cdot{\\text{e}^{-b\\cdot(x - c)}}}{(1 + \\text{e}^{-b\\cdot(x - c)})^2}`
    - Integral of the logistic model: :math:`F(x) = \\frac{a}{b}\\cdot{\\ln|\\text{e}^{b\\cdot(x - c)} + 1|}`
    - Potential x-values of the roots of the logistic model: :math:`x_{intercepts} = \\{ \\varnothing \\}`
    - Potential x-values of the maxima of the logistic model: :math:`x_{maxima} = \\{ \\varnothing \\}`
    - Potential x-values of the minima of the logistic model: :math:`x_{minima} = \\{ \\varnothing \\}`
    - Potential x-values of the inflection points of the logistic model: :math:`x_{inflections} = \\{ c \\}`
    - Accumulatation of the logistic model over its range: :math:`A_{range} = \\int_{X_{min}}^{X_{max}} f(x) \\,dx`
    - Accumulatation of the logistic model over its interquartile range: :math:`A_{iqr} = \\int_{X_{Q1}}^{X_{Q3}} f(x) \\,dx`
    - Average rate of change of the logistic model over its range: :math:`m_{range} = \\frac{f(X_{max}) - f(X_{min})}{X_{max} - X_{min}}`
    - Potential x-values at which the logistic model's instantaneous rate of change equals its average rate of change over its range: :math:`x_{m,range} = \\{ c + \\frac{1}{b}\\cdot{\\ln(2m_{range})} - \\frac{1}{b}\\cdot{\\ln\\left(ab - 2m_{range} - \\sqrt{(2m_{range} - ab)^2 - 4m_{range}^2}\\right)}, \\\\ c + \\frac{1}{b}\\cdot{\\ln(2m_{range})} - \\frac{1}{b}\\cdot{\\ln\\left(ab - 2m_{range} + \\sqrt{(2m_{range} - ab)^2 - 4m_{range}^2}\\right)} \\}`
    - Average value of the logistic model over its range: :math:`v_{range} = \\frac{1}{X_{max} - X_{min}}\\cdot{A_{range}}`
    - Potential x-values at which the logistic model's value equals its average value over its range: :math:`x_{v,range} = \\{ c - \\frac{1}{b}\\cdot{\\ln(\\frac{a}{v_{range}} - 1)} \\}`
    - Average rate of change of the logistic model over its interquartile range: :math:`m_{iqr} = \\frac{f(X_{Q3}) - f(X_{Q1})}{X_{Q3} - X_{Q1}}`
    - Potential x-values at which the logistic model's instantaneous rate of change equals its average rate of change over its interquartile range: :math:`x_{m,iqr} = \\{ c + \\frac{1}{b}\\cdot{\\ln(2m_{iqr})} - \\frac{1}{b}\\cdot{\\ln\\left(ab - 2m_{iqr} - \\sqrt{(2m_{iqr} - ab)^2 - 4m_{iqr}^2}\\right)}, \\\\ c + \\frac{1}{b}\\cdot{\\ln(2m_{iqr})} - \\frac{1}{b}\\cdot{\\ln\\left(ab - 2m_{iqr} + \\sqrt{(2m_{iqr} - ab)^2 - 4m_{iqr}^2}\\right)} \\}`
    - Average value of the logistic model over its interquartile range: :math:`v_{iqr} = \\frac{1}{X_{Q3} - X_{Q1}}\\cdot{A_{iqr}}`
    - Potential x-values at which the logistic model's value equals its average value over its interquartile range: :math:`x_{v,iqr} = \\{ c - \\frac{1}{b}\\cdot{\\ln(\\frac{a}{v_{iqr}} - 1)} \\}`
    - Predicted values based on the logistic model: :math:`\\hat{y}_i = \\{ \\hat{y}_1, \\hat{y}_2, \\cdots, \\hat{y}_n \\}`
    - Residuals of the dependent variable: :math:`e_i = \\{ p_{1,y} - \\hat{y}_1, p_{2,y} - \\hat{y}_2, \\cdots, p_{n,y} - \\hat{y}_n \\}`
    - Deviations of the dependent variable: :math:`d_i = \\{ p_{1,y} - \\bar{y}, p_{2,y} - \\bar{y}, \\cdots, p_{n,y} - \\bar{y} \\}`
    - Sum of squares of residuals: :math:`SS_{res} = \\sum\\limits_{i=1}^n e_i^2`
    - Sum of squares of deviations: :math:`SS_{dev} = \\sum\\limits_{i=1}^n d_i^2`
    - Correlation coefficient for the logistic model: :math:`r = \\sqrt{1 - \\frac{SS_{res}}{SS_{dev}}}`
    - |regression_analysis|

    Examples
    --------
    Import `logistic_model` function from `regressions` library
        >>> from regressions.models.logistic import logistic_model
    Generate a logistic regression model for the data set [[1, 0.0000122], [2, 0.000247], [3, 0.004945], [4, 0.094852], [5, 1.0], [6, 1.905148], [7, 1.995055], [8, 1.999753], [9, 1.999988], [10, 1.999999]], then print its coefficients, roots, total accumulation over its interquartile range, and correlation
        >>> model_perfect = logistic_model([[1, 0.0000122], [2, 0.000247], [3, 0.004945], [4, 0.094852], [5, 1.0], [6, 1.905148], [7, 1.995055], [8, 1.999753], [9, 1.999988], [10, 1.999999]])
        >>> print(model_perfect['constants'])
        [2.0, 3.0, 5.0]
        >>> print(model_perfect['points']['roots'])
        [None]
        >>> print(model_perfect['accumulations']['iqr'])
        5.9987
        >>> print(model_perfect['correlation'])
        1.0
    Generate a logistic regression model for the data set [[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]], then print its coefficients, inflections, total accumulation over its range, and correlation
        >>> model_agnostic = logistic_model([[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]])
        >>> print(model_agnostic['constants'])
        [43.983, 0.3076, 0.9746]
        >>> print(model_agnostic['points']['inflections'])
        [[0.9746, 21.9915]]
        >>> print(model_agnostic['accumulations']['range'])
        305.9311
        >>> print(model_agnostic['correlation'])
        0.5875
    """
    # Handle input errors
    matrix_of_scalars(data, 'first')
    long_vector(data)
    positive_integer(precision)

    # Store independent and dependent variable values separately
    independent_variable = single_dimension(data, 1)
    dependent_variable = single_dimension(data, 2)

    # Determine key values for bounds
    halved_data = half_dimension(data, 1)
    dependent_lower = single_dimension(halved_data['lower'], 2)
    dependent_upper = single_dimension(halved_data['upper'], 2)
    mean_lower = mean_value(dependent_lower)
    mean_upper = mean_value(dependent_upper)
    dependent_max = max(dependent_variable)
    dependent_min = min(dependent_variable)
    dependent_range = dependent_max - dependent_min

    # Circumvent error with bounds
    if dependent_range == 0:
        dependent_range = 1
    
    # Create function to guide model generation
    def logistic_fit(variable, first_constant, second_constant, third_constant):
        evaluation = first_constant / (1 + exp(-1 * second_constant * (variable - third_constant)))
        return evaluation
    
    # Create list to store coefficients of generated equation
    solution = []
    
    # Handle normal case where values appear to increase in the set
    if mean_upper >= mean_lower:
        # Generate model
        parameters, covariance = curve_fit(logistic_fit, independent_variable, dependent_variable, bounds=[(dependent_max - dependent_range, 0, -inf), (dependent_max + dependent_range, inf, inf)])
        solution = list(parameters)
    
    # Handle case where values do not appear to increase in the set
    else:
        # Generate model with inverted negative infinity and zero values
        parameters, covariance = curve_fit(logistic_fit, independent_variable, dependent_variable, bounds=[(dependent_max - dependent_range, -inf, -inf), (dependent_max + dependent_range, 0, inf)])
        solution = list(parameters)
    
    # Eliminate zeroes from solution
    coefficients = no_zeroes(solution, precision)

    # Generate evaluations for function, derivative, and integral
    equation = logistic_equation(*coefficients, precision)
    derivative = logistic_derivatives(*coefficients, precision)['first']['evaluation']
    integral = logistic_integral(*coefficients, precision)['evaluation']

    # Determine key points of graph
    points = key_coordinates('logistic', coefficients, precision)

    # Generate values for lower and upper bounds
    five_numbers = five_number_summary(independent_variable, precision)
    min_value = five_numbers['minimum']
    max_value = five_numbers['maximum']
    q1 = five_numbers['q1']
    q3 = five_numbers['q3']

    # Calculate accumulations
    accumulated_range = accumulated_area('logistic', coefficients, min_value, max_value, precision)
    accumulated_iqr = accumulated_area('logistic', coefficients, q1, q3, precision)

    # Determine average values and their points
    averages_range = average_values('logistic', coefficients, min_value, max_value, precision)
    averages_iqr = average_values('logistic', coefficients, q1, q3, precision)

    # Create list of predicted outputs
    predicted = []
    for element in independent_variable:
        predicted.append(equation(element))
    
    # Calculate correlation coefficient for model
    accuracy = correlation_coefficient(dependent_variable, predicted, precision)

    # Package preceding results in multiple dictionaries
    evaluations = {
        'equation': equation,
        'derivative': derivative,
        'integral': integral
    }
    points = {
        'roots': points['roots'],
        'maxima': points['maxima'],
        'minima': points['minima'],
        'inflections': points['inflections']
    }
    accumulations = {
        'range': accumulated_range,
        'iqr': accumulated_iqr
    }
    averages = {
        'range': averages_range,
        'iqr': averages_iqr
    }

    # Package all dictionaries in single dictionary to return
    result = {
        'constants': coefficients,
        'evaluations': evaluations,
        'points': points,
        'accumulations': accumulations,
        'averages': averages,
        'correlation': accuracy
    }
    return result


from regressions.errors.scalars import compare_scalars, positive_integer
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.analyses import select_equations
from regressions.statistics.rounding import rounded_value
from .integrals.linear import linear_integral
from .integrals.quadratic import quadratic_integral
from .integrals.cubic import cubic_integral
from .integrals.hyperbolic import hyperbolic_integral
from .integrals.exponential import exponential_integral
from .integrals.logarithmic import logarithmic_integral
from .integrals.logistic import logistic_integral
from .integrals.sinusoidal import sinusoidal_integral

def accumulated_area(equation_type, coefficients, start, end, precision = 4):
    """
    Evaluates the definite integral between two points for a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which the definite integral must be evaluated (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients of the original function to integrate
    start : int or float
        Value of the x-coordinate of the first point to use for evaluating the definite integral
    end : int or float
        Value of the x-coordinate of the second point to use for evaluating the definite integral
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third and fourth arguments must be integers or floats
    ValueError
        Third argument must be less than or equal to fourth argument
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    area : float
        Definite integral of the original equation, evaluated between two points; if start and end values are identical, then definite integral will be zero

    See Also
    --------
    :func:`~regressions.analyses.integrals.linear.linear_integral`, :func:`~regressions.analyses.integrals.quadratic.quadratic_integral`, :func:`~regressions.analyses.integrals.cubic.cubic_integral`, :func:`~regressions.analyses.integrals.hyperbolic.hyperbolic_integral`, :func:`~regressions.analyses.integrals.exponential.exponential_integral`, :func:`~regressions.analyses.integrals.logarithmic.logarithmic_integral`, :func:`~regressions.analyses.integrals.logistic.logistic_integral`, :func:`~regressions.analyses.integrals.sinusoidal.sinusoidal_integral`

    Notes
    -----
    - Definite integral of a function: :math:`\\int_{a}^{b} f(x) \\,dx = F(b) - F(a)`
    - |definite_integral|
    - |fundamental_theorem|

    Examples
    --------
    Import `accumulated_area` function from `regressions` library
        >>> from regressions.analyses.accumulation import accumulated_area
    Evaluate the definite integral of a linear function with coefficients 2 and 3 between the end points 10 and 20
        >>> area_linear = accumulated_area('linear', [2, 3], 10, 20)
        >>> print(area_linear)
        330.0
    Evaluate the definite integral of a cubic function with coefficients 8, 6, -10, and 7 between the end points 10 and 20
        >>> area_cubic = accumulated_area('cubic', [8, 6, -10, 7], 10, 20)
        >>> print(area_cubic)
        312570.0
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    compare_scalars(start, end, 'third', 'fourth')
    positive_integer(precision)

    # Create indefinite integral based on equation type
    integral = lambda x : x
    if equation_type == 'linear':
        integral = linear_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'quadratic':
        integral = quadratic_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'cubic':
        integral = cubic_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'hyperbolic':
        integral = hyperbolic_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'exponential':
        integral = exponential_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'logarithmic':
        integral = logarithmic_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'logistic':
        integral = logistic_integral(*coefficients, precision)['evaluation']
    elif equation_type == 'sinusoidal':
        integral = sinusoidal_integral(*coefficients, precision)['evaluation']
    
    # Evaluate definite integral
    area = integral(end) - integral(start)

    # Round evaluation
    result = rounded_value(area, precision)
    return result


from math import floor
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from regressions.errors.matrices import matrix_of_scalars
from .sort import sorted_list, sorted_dimension

def partition(data):
    """
    Splits an unsorted data set into two unsorted data sets, each containing the same amount of elements, with elements allocated soley according to where they happen to appear in the original data set (in sets with an odd amount of elements, the median is not included in either half)

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze

    Returns
    -------
    sections['upper'] : list of int or float
        List of all elements from the upper half of a data set
    sections['lower'] : list of int or float
        List of all elements from the lower half of a data set

    See Also
    --------
    :func:`~regressions.vectors.dimension.single_dimension`, :func:`~regressions.vectors.unify.unite_vectors`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - For sets with an odd amount of numbers:

        - Lower section: :math:`a_{lower} = \\{ a_1, a_2, \\cdots, a_{\\lfloor n/2 \\rfloor} \\}`
        - Upper section: :math:`a_{upper} = \\{ a_{\\lceil n/2 \\rceil}, a_{\\lceil n/2 \\rceil + 1}, \\cdots, a_n \\}`
    
    - For sets with an even amount of numbers:

        - Lower section: :math:`a_{lower} = \\{ a_1, a_2, \\cdots, a_{n/2} \\}`
        - Upper section: :math:`a_{upper} = \\{ a_{n/2 + 1}, a_{n/2 + 2}, \\cdots, a_n \\}`

    Examples
    --------
    Import `partition` function from `regressions` library
        >>> from regressions.statistics.halve import partition
    Determine the upper half of the set [5, 2, 9, 8]
        >>> sections_short = partition([5, 2, 9, 8])
        >>> print(sections_short['upper'])
        [9, 8]
    Determine the lower half of the set [11, 3, 52, 25, 21, 25, 6]
        >>> sections_long = partition([11, 3, 52, 25, 21, 25, 6])
        >>> print(sections_long['lower'])
        [11, 3, 52]
    """
    # Determine length of input
    length = len(data)

    # Create intermediary lists
    upper = []
    lower = []

    # Handle an even amount of elements
    if length % 2 == 0:
        index = int(length / 2)
        upper = data[index:]
        lower = data[:index]
    
    # Handle an odd amount of elements
    else:
        index = int(floor(length / 2))
        upper = data[index + 1:]
        lower = data[:index]
    
    # Package both lists in single dictionary to return
    result = {
        'upper': upper,
        'lower': lower
    }
    return result

def half(data):
    """
    Splits an unsorted data set into two sorted data sets, each containing the same amount of elements (in sets with an odd amount of elements, the median is not included in either half)

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze
    
    Raises
    ------
    TypeError
        Argument must be a 1-dimensional list
    TypeError
        Elements of argument must be integers or floats

    Returns
    -------
    sections['upper'] : list of int or float
        List of all elements from the upper half of a sorted data set
    sections['lower'] : list of int or float
        List of all elements from the lower half of a sorted data set

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Sorted version of set: :math:`A_i = ( A_1, A_2, \\cdots, A_n )`

        - For all terms in :math:`A_i`: :math:`A_{n-1} \\leq A_n`

    - For sets with an odd amount of numbers:

        - Lower section: :math:`A_{lower} = ( A_1, A_2, \\cdots, A_{\\lfloor n/2 \\rfloor} )`
        - Upper section: :math:`A_{upper} = ( A_{\\lceil n/2 \\rceil}, A_{\\lceil n/2 \\rceil + 1}, \\cdots, A_n )`
    
    - For sets with an even amount of numbers:

        - Lower section: :math:`A_{lower} = ( A_1, A_2, \\cdots, A_{n/2} )`
        - Upper section: :math:`A_{upper} = ( A_{n/2 + 1}, A_{n/2 + 2}, \\cdots, A_n )`

    Examples
    --------
    Import `half` function from `regressions` library
        >>> from regressions.statistics.halve import half
    Determine the sorted upper half of the set [5, 2, 9, 8]
        >>> sections_short = half([5, 2, 9, 8])
        >>> print(sections_short['upper'])
        [8, 9]
    Determine the sorted lower half of the set [11, 3, 52, 25, 21, 25, 6]
        >>> sections_long = half([11, 3, 52, 25, 21, 25, 6])
        >>> print(sections_long['lower'])
        [3, 6, 11]
    """
    # Handle input errors
    vector_of_scalars(data)

    # Sort input
    sorted_data = sorted_list(data)

    # Partition sorted data
    result = partition(sorted_data)
    return result

def half_dimension(data, dimension = 1):
    """
    Splits an unsorted 2-dimensional data set into two sorted 2-dimensional data sets, each containing the same amount of elements, in which the sorting occurs based on the elements of the nested lists indicated by the dimension parameter (in sets with an odd amount of elements, the median is not included in either half)

    Parameters
    ----------
    data : list of lists of int or float
        List of lists of numbers to analyze
    dimension : int, default=1
        Number indicating by which element of the nested lists to sort
    
    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    sections['upper'] : list of lists of int or float
        List of all elements from the upper half of a data set, sorted according to the elements occupying a provided position
    sections['lower'] : list of lists of int or float
        List of all elements from the lower half of a data set, sorted according to the elements occupying a provided position

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_dimension`, :func:`~regressions.vectors.dimension.single_dimension`

    Notes
    -----
    - Set of ordered pairs of numbers: :math:`a_i = \\{ ( a_{1,1}, a_{1,2}, \\cdots, a_{1,j}, a_{1,n} ), ( a_{2,1}, a_{2,2}, \\cdots, a_{2,j}, a_{2,n} ), \\cdots, \\\\ ( a_{m,1}, a_{m,2}, \\cdots, a_{m,j}, a_{m,n} ) \\}`
    - Sorted version of set according to the values in the :math:`j`\ th position: :math:`A_i = ( ( A_{1,1}, A_{1,2}, \\cdots, A_{1,j}, A_{1,n} ), ( A_{2,1}, A_{2,2}, \\cdots, A_{2,j}, A_{2,n} ), \\cdots, \\\\ ( A_{m,1}, A_{m,2}, \\cdots, A_{m,j}, A_{m,n} ) )`

        - For all terms in :math:`A_i`: :math:`A_{n-1,j} \\leq A_{n,j}`
    
    - For sets with an odd amount of ordered pairs:

        - Lower section: :math:`A_{lower} = ( ( A_{1,1}, A_{1,2}, \\cdots, A_{1,j}, A_{1,n} ), ( A_{2,1}, A_{2,2}, \\cdots, A_{2,j}, A_{2,n} ), \\cdots, \\\\ ( A_{\\lfloor m/2 \\rfloor,1}, A_{\\lfloor m/2 \\rfloor,2}, \\cdots, A_{\\lfloor m/2 \\rfloor,j}, A_{\\lfloor m/2 \\rfloor,n} ) )`
        - Upper section: :math:`A_{upper} = ( ( A_{\\lceil m/2 \\rceil,1}, A_{\\lceil m/2 \\rceil,2}, \\cdots, A_{\\lceil m/2 \\rceil,j}, A_{\\lceil m/2 \\rceil,n} ), ( A_{\\lceil m/2 \\rceil + 1,1}, A_{\\lceil m/2 \\rceil + 1,2}, \\cdots, \\\\ A_{\\lceil m/2 \\rceil + 1,j}, A_{\\lceil m/2 \\rceil + 1,n} ), \\cdots, ( A_{m,1}, A_{m,2}, \\cdots, A_{m,j}, A_{m,n} ) )`
    
    - For sets with an even amount of ordered pairs:

        - Lower section: :math:`A_{lower} = ( ( A_{1,1}, A_{1,2}, \\cdots, A_{1,j}, A_{1,n} ), ( A_{2,1}, A_{2,2}, \\cdots, A_{2,j}, A_{2,n} ), \\cdots, \\\\ ( A_{m/2,1}, A_{m/2,2}, \\cdots, A_{m/2,j}, A_{m/2,n} ) )`
        - Upper section: :math:`A_{upper} = ( ( A_{m/2 + 1,1}, A_{m/2 + 1,2}, \\cdots, A_{m/2 + 1,j}, A_{m/2 + 1,n} ), ( A_{m/2 + 2,1}, A_{m/2 + 2,2}, \\cdots, \\\\ A_{m/2 + 2,j}, A_{m/2 + 2,n} ), \\cdots, ( A_{m,1}, A_{m,2}, \\cdots, A_{m,j}, A_{m,n} ) )`

    Examples
    --------
    Import `half_dimension` function from `regressions` library
        >>> from regressions.statistics.halve import half_dimension
    Determine the upper half of the set [[3, 7, 1], [1, 8, 11], [6, 6, 6], [2, 15, 3], [10, 5, 9]] based on the second dimension
        >>> sections_2d = half_dimension([[3, 7, 1], [1, 8, 11], [6, 6, 6], [2, 15, 3], [10, 5, 9]], 2)
        >>> print(sections_2d['upper'])
        [[1, 8, 11], [2, 15, 3]]
    Determine the lower half of the set [[3, 7, 1], [1, 8, 11], [6, 6, 6], [2, 15, 3], [10, 5, 9]] based on the third dimension
        >>> sections_3d = half_dimension([[3, 7, 1], [1, 8, 11], [6, 6, 6], [2, 15, 3], [10, 5, 9]], 3)
        >>> print(sections_3d['lower'])
        [[3, 7, 1], [2, 15, 3]]
    """
    # Handle input errors
    matrix_of_scalars(data, 'first')
    positive_integer(dimension)

    # Sort input according to a provided dimension
    sorted_data = sorted_dimension(data, dimension)

    # Partition sorted input
    result = partition(sorted_data)
    return result


from regressions.errors.scalars import four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def cubic_derivatives(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Calculates the first and second derivatives of a cubic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the cubic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the quadratic term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Coefficient of the linear term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Coefficient of the constant term of the original cubic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    derivatives['first']['constants'] : list of float
        Coefficients of the resultant first derivative
    derivatives['first']['evaluation'] : func
        Function for evaluating the resultant first derivative at any float or integer argument
    derivatives['second']['constants'] : list of float
        Coefficients of the resultant second derivative
    derivatives['second']['evaluation'] : func
        Function for evaluating the resultant second derivative at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.cubic.cubic_equation`, :func:`~regressions.analyses.integrals.cubic.cubic_integral`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.models.cubic.cubic_model`

    Notes
    -----
    - Standard form of a cubic function: :math:`f(x) = a\\cdot{x^3} + b\\cdot{x^2} + c\\cdot{x} + d`
    - First derivative of a cubic function: :math:`f'(x) = 3a\\cdot{x^2} + 2b\\cdot{x} + c`
    - Second derivative of a cubic function: :math:`f''(x) = 6a\\cdot{x} + 2b`
    - |differentiation_formulas|

    Examples
    --------
    Import `cubic_derivatives` function from `regressions` library
        >>> from regressions.analyses.derivatives.cubic import cubic_derivatives
    Generate the derivatives of a cubic function with coefficients 2, 3, 5, and 7, then display the coefficients of its first and second derivatives
        >>> derivatives_constants = cubic_derivatives(2, 3, 5, 7)
        >>> print(derivatives_constants['first']['constants'])
        [6.0, 6.0, 5.0]
        >>> print(derivatives_constants['second']['constants'])
        [12.0, 6.0]
    Generate the derivatives of a cubic function with coefficients 7, -5, -3, and 2, then evaluate its first and second derivatives at 10
        >>> derivatives_evaluation = cubic_derivatives(7, -5, -3, 2)
        >>> print(derivatives_evaluation['first']['evaluation'](10))
        1997.0
        >>> print(derivatives_evaluation['second']['evaluation'](10))
        410.0
    Generate the derivatives of a cubic function with all inputs set to 0, then display the coefficients of its first and second derivatives
        >>> derivatives_zeroes = cubic_derivatives(0, 0, 0, 0)
        >>> print(derivatives_zeroes['first']['constants'])
        [0.0003, 0.0002, 0.0001]
        >>> print(derivatives_zeroes['second']['constants'])
        [0.0006, 0.0002]
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)

    # Create first derivative
    first_constants = [3 * coefficients[0], 2 * coefficients[1], coefficients[2]]
    def first_derivative(variable):
        evaluation = first_constants[0] * variable**2 + first_constants[1] * variable + first_constants[2]
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    first_dictionary = {
        'constants': first_constants,
        'evaluation': first_derivative
    }

    # Create second derivative
    second_constants = [2 * first_constants[0], first_constants[1]]
    def second_derivative(variable):
        evaluation = second_constants[0] * variable + second_constants[1]
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    second_dictionary = {
        'constants': second_constants,
        'evaluation': second_derivative
    }
    
    # Package both derivatives in single dictionary
    results = {
        'first': first_dictionary,
        'second': second_dictionary
    }
    return results


from regressions.errors.analyses import select_equations
from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from .intervals import sign_chart

def inflection_points(equation_type, coefficients, precision = 4):
    """
    Calculates the inflection points of a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which inflections must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points : list of float or str
        Values of the x-coordinates at which the original function has an inflection point; if the function is sinusoidal, then only five results within a two-period interval will be listed, but a general form will also be included; if the function has no inflection points, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.points.key_coordinates`

    Notes
    -----
    - Critical points for the second derivative of a function: :math:`c_i = \\{ c_1, c_2, c_3,  \\cdots, c_{n-1}, c_n \\}`
    - X-coordinates of the inflections of the function: :math:`x_{infl} = \\{ x \\mid x \\in c_i, \\left( f''(\\frac{c_{j-1} + c_j}{2}) < 0 \\cap f''(\\frac{c_j + c_{j+1}}{2}) > 0 \\right) \\\\ \\cup \\left( f''(\\frac{c_{j-1} + c_j}{2}) > 0 \\cap f''(\\frac{c_j + c_{j+1}}{2}) < 0 \\right) \\}`
    - |inflection_points|

    Examples
    --------
    Import `inflection_points` function from `regressions` library
        >>> from regressions.analyses.inflections import inflection_points
    Calculate the inflection points of a cubic functions with coefficients 1, -15, 63, and -7
        >>> points_cubic = inflection_points('cubic', [1, -15, 63, -7])
        >>> print(points_cubic)
        [5.0]
    Calculate the inflection points of a sinusoidal functions with coefficients 2, 3, 5, and 7
        >>> points_sinusoidal = inflection_points('sinusoidal', [2, 3, 5, 7])
        >>> print(points_sinusoidal)
        [5.0, 6.0472, 7.0944, 8.1416, 9.1888, '5.0 + 1.0472k']
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    positive_integer(precision)

    # Create sign chart
    intervals_set = sign_chart(equation_type, coefficients, 2, precision)

    # Handle hyperbolic case
    if equation_type == 'hyperbolic':
        result = [None]
        return result
    
    # Determine inflections
    result = []
    for i in range(len(intervals_set)):
        try:
            if (intervals_set[i] == 'positive' and intervals_set[i + 2] == 'negative') or (intervals_set[i] == 'negative' and intervals_set[i + 2] == 'positive'):
                result.append(intervals_set[i + 1])
        except IndexError:
            pass
    
    # Handle sinusoidal case
    if equation_type == 'sinusoidal':
        result.append(intervals_set[-1])

    # Handle no inflections
    if len(result) == 0:
        result.append(None)
    return result


from regressions.errors.analyses import select_equations, select_points
from regressions.errors.scalars import scalar_value, compare_scalars, positive_integer
from regressions.errors.vectors import vector_of_scalars, allow_none_vector
from regressions.errors.matrices import allow_none_matrix, allow_vector_matrix
from regressions.statistics.rounding import rounded_value, rounded_list
from regressions.statistics.sort import sorted_list, sorted_strings
from regressions.statistics.ranges import shift_into_range
from regressions.vectors.unify import unite_vectors
from regressions.vectors.separate import separate_elements
from regressions.vectors.generate import generate_elements
from .equations.linear import linear_equation
from .equations.quadratic import quadratic_equation
from .equations.cubic import cubic_equation
from .equations.hyperbolic import hyperbolic_equation
from .equations.exponential import exponential_equation
from .equations.logarithmic import logarithmic_equation
from .equations.logistic import logistic_equation
from .equations.sinusoidal import sinusoidal_equation
from .intercepts import intercept_points
from .extrema import extrema_points
from .inflections import inflection_points

def coordinate_pairs(equation_type, coefficients, inputs, point_type = 'point', precision = 4):
    """
    Creates a list of coordinate pairs from a set of inputs

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which coordinate pairs must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    inputs : list of int or float or str
        X-coordinates to use to generate the y-coordinates for each coordinate pair
    point_type : str, default='point'
        Name of the type of point that describes all points which must be generated (e.g., 'intercepts', 'maxima')
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    TypeError
        Third argument must be a 1-dimensional list containing elements that are integers, floats, strings, or None
    ValueError
        Fourth argument must be either 'point', 'intercepts', 'maxima', 'minima', or 'inflections'
    ValueError
        Last argument must be a positive integer
        
    Returns
    -------
    points : list of float or str
        List containing lists of coordinate pairs, in which the second element of the inner lists are floats and the first elements of the inner lists are either floats or strings (the latter for general forms); may return a list of None if inputs list contained None

    See Also
    --------
    :func:`~regressions.vectors.generate.generate_elements`, :func:`~regressions.vectors.unify.unite_vectors`

    Notes
    -----
    - Set of x-coordinates of points: :math:`x_i = \\{ x_1, x_2, \\cdots, x_n \\}`
    - Set of y-coordinates of points: :math:`y_i = \\{ y_1, y_2, \\cdots, y_n \\}`
    - Set of coordinate pairs of points: :math:`p_i = \\{ (x_1, y_1), (x_2, y_2), \\cdots, (x_n, y_n) \\}`

    Examples
    --------
    Import `coordinate_pairs` function from `regressions` library
        >>> from regressions.analyses.points import coordinate_pairs
    Generate a list of coordinate pairs for a cubic function with coefficients 2, 3, 5, and 7 based off x-coordinates of 1, 2, 3, and 4
        >>> points_cubic = coordinate_pairs('cubic', [2, 3, 5, 7], [1, 2, 3, 4])
        >>> print(points_cubic)
        [[1.0, 17.0], [2.0, 45.0], [3.0, 103.0], [4.0, 203.0]]
    Generate a list of coordinate pairs for a sinusoidal function with coefficients 2, 3, 5, and 7 based off x-coordinates of 1, 2, 3, and 4
        >>> points_sinusoidal = coordinate_pairs('sinusoidal', [2, 3, 5, 7], [1, 2, 3, 4])
        >>> print(points_sinusoidal)
        [[1.0, 8.0731], [2.0, 6.1758], [3.0, 7.5588], [4.0, 6.7178]]
    Generate a list of coordinate pairs for a quadratic function with coefficients 1, -5, and 6 based off x-coordinates of 2 and 3 (given that the resultant coordinates will be x-intercepts)
        >>> points_quadratic = coordinate_pairs('quadratic', [1, -5, 6], [2, 3], 'intercepts')
        >>> print(points_quadratic)
        [[2.0, 0.0], [3.0, 0.0]]
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    allow_none_vector(inputs, 'third')
    select_points(point_type, 'fourth')
    positive_integer(precision)

    # Create equations for evaluating inputs (based on equation type)
    equation = lambda x : x
    if equation_type == 'linear':
        equation = linear_equation(*coefficients, precision)
    elif equation_type == 'quadratic':
        equation = quadratic_equation(*coefficients, precision)
    elif equation_type == 'cubic':
        equation = cubic_equation(*coefficients, precision)
    elif equation_type == 'hyperbolic':
        equation = hyperbolic_equation(*coefficients, precision)
    elif equation_type == 'exponential':
        equation = exponential_equation(*coefficients, precision)
    elif equation_type == 'logarithmic':
        equation = logarithmic_equation(*coefficients, precision)
    elif equation_type == 'logistic':
        equation = logistic_equation(*coefficients, precision)
    elif equation_type == 'sinusoidal':
        equation = sinusoidal_equation(*coefficients, precision)

    # Round inputs
    rounded_inputs = []
    for point in inputs:
        if isinstance(point, (int, float)):
            rounded_inputs.append(rounded_value(point, precision))
        else:
            rounded_inputs.append(point)

    # Create empty lists
    outputs = []
    coordinates = []
    
    # Handle no points
    if rounded_inputs[0] == None:
        coordinates.append(None)
    
    # Fill outputs list with output value at each input
    else:
        for value in rounded_inputs:
            # Circumvent inaccurate rounding
            if point_type == 'intercepts':
                outputs.append(0.0)
            
            # Evaluate function at inputs
            else:
                # Evaluate numerical inputs
                if isinstance(value, (int, float)):
                    output = equation(value)
                    rounded_output = rounded_value(output, precision)
                    outputs.append(rounded_output)
                
                # Handle non-numerical inputs
                else:
                    outputs.append(outputs[0])

        # Unite inputs and outputs for maxima into single list
        coordinates.extend(unite_vectors(rounded_inputs, outputs))
    
    # Return final coordinate pairs
    return coordinates

def key_coordinates(equation_type, coefficients, precision = 4):
    """
    Calculates the key points of a specific function

    Parameters
    ----------
    equation_type : str
        Name of the type of function for which key points must be determined (e.g., 'linear', 'quadratic')
    coefficients : list of int or float
        Coefficients to use to generate the equation to investigate
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    ValueError
        First argument must be either 'linear', 'quadratic', 'cubic', 'hyperbolic', 'exponential', 'logarithmic', 'logistic', or 'sinusoidal'
    TypeError
        Second argument must be a 1-dimensional list containing elements that are integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    points['roots'] : list of float or str
        List containing two-element lists for each point; first elements of those lists will be the value of the x-coordinate at which the original function has a root; second elements of those lists will be 0; if the function is sinusoidal, then only the initial results within a four-period interval will be listed, but general forms will also be included; if the function has no roots, then it will return a list of `None`
    points['maxima'] : list of float or str
        List containing two-element lists for each point; first elements of those lists will be the value of the x-coordinate at which the original function has a relative maximum; second elements of those lists will be the y-coordinate of that maximum; if the function is sinusoidal, then only the initial results within a two-period interval will be listed, but a general form will also be included; if the function has no maxima, then it will return a list of `None`
    points['minima'] : list of float or str
        List containing two-element lists for each point; first elements of those lists will be the value of the x-coordinate at which the original function has a relative minimum; second elements of those lists will be the y-coordinate of that minimum; if the function is sinusoidal, then only the initial results within a two-period interval will be listed, but a general form will also be included; if the function has no minima, then it will return a list of `None`
    points['inflections'] : list of float or str
        List containing two-element lists for each point; first elements of those lists will be the value of the x-coordinate at which the original function has an inflection; second elements of those lists will be the y-coordinate of that inflection; if the function is sinusoidal, then only the initial results within a two-period interval will be listed, but a general form will also be included; if the function has no inflection points, then it will return a list of `None`

    See Also
    --------
    - Roots for key functions: :func:`~regressions.analyses.roots.linear.linear_roots`, :func:`~regressions.analyses.roots.quadratic.quadratic_roots`, :func:`~regressions.analyses.roots.cubic.cubic_roots`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.analyses.roots.exponential.exponential_roots`, :func:`~regressions.analyses.roots.logarithmic.logarithmic_roots`, :func:`~regressions.analyses.roots.logistic.logistic_roots`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`
    - Graphical analysis: :func:`~regressions.analyses.criticals.critical_points`, :func:`~regressions.analyses.intervals.sign_chart`, :func:`~regressions.analyses.maxima.maxima_points`, :func:`~regressions.analyses.minima.minima_points`, :func:`~regressions.analyses.extrema.extrema_points`, :func:`~regressions.analyses.inflections.inflection_points`

    Notes
    -----
    - Key points include x-intercepts, maxima, minima, and points of inflection
    - |intercepts|
    - |extrema|
    - |inflections|

    Examples
    --------
    Import `key_coordinates` function from `regressions` library
        >>> from regressions.analyses.points import key_coordinates
    Calculate the key points of a cubic function with coefficients 1, -15, 63, and -7
        >>> points_cubic = key_coordinates('cubic', [1, -15, 63, -7])
        >>> print(points_cubic['roots'])
        [[0.1142, 0.0]]
        >>> print(points_cubic['maxima'])
        [[3.0, 74.0]]
        >>> print(points_cubic['minima'])
        [[7.0, 42.0]]
        >>> print(points_cubic['inflections'])
        [[5.0, 58.0]]
    Calculate the key points of a sinusoidal function with coefficients 2, 3, 5, and 1
        >>> points_sinusoidal = key_coordinates('sinusoidal', [2, 3, 5, 1])
        >>> print(points_sinusoidal['roots'])
        [[4.8255, 0.0], [6.2217, 0.0], [6.9199, 0.0], [8.3161, 0.0], [9.0143, 0.0], [10.4105, 0.0], [11.1087, 0.0], [12.5049, 0.0], [13.203, 0.0], [14.5993, 0.0], ['4.8255 + 2.0944k', 0.0], ['6.2217 + 2.0944k', 0.0]]
        >>> print(points_sinusoidal['maxima'])
        [[5.5236, 3.0], [7.618, 3.0], [9.7124, 3.0], ['5.5236 + 2.0944k', 3.0]]
        >>> print(points_sinusoidal['minima'])
        [[6.5708, -1.0], [8.6652, -1.0], ['6.5708 + 2.0944k', -1.0]]
        >>> print(points_sinusoidal['inflections'])
        [[5.0, 1.0], [6.0472, 1.0], [7.0944, 1.0], [8.1416, 1.0], [9.1888, 1.0001], ['5.0 + 1.0472k', 1.0]]
    """
    # Handle input errors
    select_equations(equation_type)
    vector_of_scalars(coefficients, 'second')
    positive_integer(precision)
    
    # Create lists of inputs
    intercepts_inputs = intercept_points(equation_type, coefficients, precision)
    extrema_inputs = extrema_points(equation_type, coefficients, precision)
    maxima_inputs = extrema_inputs['maxima']
    minima_inputs = extrema_inputs['minima']
    inflections_inputs = inflection_points(equation_type, coefficients, precision)

    # Generate coordinate pairs for all x-intercepts
    intercepts_coordinates = coordinate_pairs(equation_type, coefficients, intercepts_inputs, 'intercepts', precision)
    
    # Generate coordinate pairs for all maxima
    maxima_coordinates = coordinate_pairs(equation_type, coefficients, maxima_inputs, 'maxima', precision)
    
    # Generate coordinate pairs for all minima
    minima_coordinates = coordinate_pairs(equation_type, coefficients, minima_inputs, 'minima', precision)
    
    # Generate coordinate pairs for all points of inflection
    inflections_coordinates = coordinate_pairs(equation_type, coefficients, inflections_inputs, 'inflections', precision)
    
    # Create dictionary to return
    result = {
        'roots': intercepts_coordinates,
        'maxima': maxima_coordinates,
        'minima': minima_coordinates,
        'inflections': inflections_coordinates
    }
    return result

def points_within_range(points, start, end):
    """
    Eliminates all values from a set of points that fall below a lower bound or above an upper bound

    Parameters
    ----------
    points : list of int or float or str
        Set of points to narrow down to only those within a certain range
    start : int or float
        Lower bound of range into which the initial value must be adjusted (final value should be greater than or equal to start)
    end : int or float
        Upper bound of range into which the initial value must be adjusted (final value should be less than or equal to end)

    Raises
    ------
    TypeError
        First argument must be a 1-dimensional list containing elements that are integers, floats, strings, or None
    TypeError
        Second and third arguments must be integers or floats
    ValueError
        Second argument must be less than or equal to third argument

    Returns
    -------
    selected_points : list of int or float or str
        List of all values from original list that fall within specified range; may return a list of None if no points from the original list fall within range or if original list only contained None

    See Also
    --------
    :func:`~regressions.vectors.separate.separate_elements`, :func:`~regressions.statistics.ranges.shift_into_range`, :func:`~regressions.analyses.mean_values.mean_values_derivative`, :func:`~regressions.analyses.mean_values.mean_values_integral`

    Notes
    -----
    - Initial set of points: :math:`p_i = \\{ p_1, p_2, \\cdots, p_n \\}`
    - Lower bound of range: :math:`b_l`
    - Upper bound of range: :math:`b_u`
    - Adjusted set of points within range: :math:`r_i = \\{ r \\mid r \\in p_i, r \\geq b_l, r \\leq b_u \\}`

    Examples
    --------
    Import `points_within_range` function from `regressions` library
        >>> from regressions.analyses.points import points_within_range
    Eliminate all points above 19 or below 6 in the set [1, 5, 6, 7, 18, 20, 50, 127]
        >>> selected_points_int = points_within_range([1, 5, 6, 7, 18, 20, 50, 127], 6, 19)
        >>> print(selected_points_int)
        [6, 7, 18]
    Eliminate all points above 243.7821 or below 198.1735 in the set [542.1234, 237.9109, -129.3214, 199.4321, 129.3214]
        >>> selected_points_float = points_within_range([542.1234, 237.9109, -129.3214, 199.4321, 129.3214], 198.1735, 243.7821)
        >>> print(selected_points_float)
        [237.9109, 199.4321]
    """
    # Handle input errors
    allow_none_vector(points, 'first')
    compare_scalars(start, end, 'second', 'third')

    # Separate numerical results from string results
    separated_results = separate_elements(points)
    numerical_results = separated_results['numerical']
    other_results = separated_results['other']

    # Eliminate numerical results outside of range
    selected_results = [x for x in numerical_results if x >= start and x <= end]

    # Create list to return
    final_results = []

    # Handle no results
    if not selected_results and not other_results:
        final_results.append(None)
    
    # Handle general case
    else:
        final_results.extend(selected_results + other_results)
    
    # Return results
    return final_results

def shifted_points_within_range(points, minimum, maximum, precision = 4):
    # Handle input errors
    allow_vector_matrix(points, 'first')
    compare_scalars(minimum, maximum, 'second', 'third')
    positive_integer(precision)

    # Grab general points
    general_points = []
    for point in points:
        # Handle coordinate pairs
        if isinstance(point, list):
            if isinstance(point[0], str):
                general_points.append(point[0])
        
        # Handle single coordinates
        else:
            if isinstance(point, str):
                general_points.append(point)
    
    # Generate options for inputs
    optional_points = []
    for point in general_points:
        # Grab initial value and periodic unit
        initial_value_index = point.find(' + ')
        initial_value = float(point[:initial_value_index])
        periodic_unit_index = initial_value_index + 3
        periodic_unit = float(point[periodic_unit_index:-1])
        
        # Increase or decrease initial value to fit into range
        alternative_initial_value = shift_into_range(initial_value, periodic_unit, minimum, maximum)
        
        # Generate additional values within range
        generated_elements = generate_elements(alternative_initial_value, periodic_unit, precision)
        optional_points += generated_elements
    
    # Separate numerical inputs from string inputs
    separated_points = separate_elements(optional_points)
    numerical_points = separated_points['numerical']
    other_points = separated_points['other']

    # Sort numerical inputs
    sorted_points = sorted_list(numerical_points)

    # Reduce numerical inputs to within a given range
    selected_points = [x for x in sorted_points if x >= minimum and x <= maximum]
    
    # Round numerical inputs
    rounded_points = rounded_list(selected_points, precision)
    
    # Sort string inputs
    sorted_other_points = sorted_strings(other_points)
    
    # Combine numerical and string inputs
    result = rounded_points + sorted_other_points
    return result

def shifted_coordinates_within_range(coordinates, minimum, maximum, precision = 4):
    # Handle input errors
    allow_none_matrix(coordinates, 'first')
    compare_scalars(minimum, maximum, 'second', 'third')
    positive_integer(precision)

    # Create list to return
    result = []

    # Handle general case
    if coordinates[0] is not None:
        # Generate inputs
        input_points = shifted_points_within_range(coordinates, minimum, maximum, precision)
        
        # Generate outputs
        output_points = []
        for point in input_points:
            output_points.append(coordinates[0][1])
        
        # Unite inputs and outputs into single list
        result = unite_vectors(input_points, output_points)
    
    # Handle no points
    else:
        result = coordinates
    
    # Return result
    return result


from regressions.errors.scalars import two_scalars, positive_integer
from regressions.statistics.sort import sorted_list
from regressions.statistics.rounding import rounded_value, rounded_list

def generate_elements(initial_value, periodic_unit, precision = 4):
    """
    Generates a vector containing an initial numerical value, four additional numerical values created by incrementing the initial value by a periodic unit four times, and a string value for the general form of all values in the vector, made up of a multiple of the periodic unit added to the initial value

    Parameters
    ----------
    initial_value : int or float
        Starting value to adjust to fit into a range
    periodic_unit : int or float
        Unit by which the initial value should be incrementally increased or decreased to fit into a range
    precision : int, default=4
        Upper bound of range into which the initial value must be adjusted (final value should be less than or equal to maximum)

    Raises
    ------
    TypeError
        First and second arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    generated_vector : list of float
        Vector containing five numerical values, each a set incremenent apart from one another, and a string value representing the general form of all numerical elements in the vector as well as any additional numerical elements that could be generated from it in the future

    See Also
    --------
    :func:`~regressions.statistics.ranges.shift_into_range`, :func:`~regressions.analyses.points.shifted_points_within_range`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`

    Notes
    -----
    - Initial value: :math:`v_i`
    - Periodic unit: :math:`\\lambda`
    - Set of all values derived from initial value and periodic unit: :math:`g = \\{ v \\mid v = v_i + \\lambda\\cdot{k} \\}`

        - :math:`k \\in \\mathbb{Z}`

    Examples
    --------
    Import `generate_elements` function from `regressions` library
        >>> from regressions.vectors.generate import generate_elements
    Generate a vector of elements based off an initial value of 3 and a periodic unit of 2
        >>> generated_vector_int = generate_elements(3, 2)
        >>> print(generated_vector_int)
        [3.0, 5.0, 7.0, 9.0, 11.0, '3.0 + 2.0k']
    Generate a vector of elements based off an initial value of 17.23 and a periodic unit of 5.89
        >>> generated_vector_float = generate_elements(17.23, 5.89)
        >>> print(generated_vector_float)
        [17.23, 23.12, 29.01, 34.9, 40.79, '17.23 + 5.89k']
    """
    # Handle input errors
    two_scalars(initial_value, periodic_unit)
    positive_integer(precision)

    # Generate values from inputs
    first_value = initial_value + 1 * periodic_unit
    second_value = initial_value + 2 * periodic_unit
    third_value = initial_value + 3 * periodic_unit
    fourth_value = initial_value + 4 * periodic_unit

    # Store values in list
    values = [initial_value, first_value, second_value, third_value, fourth_value]

    # Sort values
    sorted_values = sorted_list(values)

    # Round values
    rounded_values = rounded_list(sorted_values, precision)

    # Create general form
    rounded_periodic_unit = rounded_value(periodic_unit, precision)
    rounded_initial_value = rounded_value(initial_value, precision)
    general_form = str(rounded_initial_value) + ' + ' + str(rounded_periodic_unit) + 'k'

    # Store numerical values and general form in single list
    results = [*rounded_values, general_form]
    return results


from .positions import argument_position

def confirm_vector(vector, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(vector, list) or isinstance(vector[0], list):
        raise TypeError(f'{identifier.capitalize()} must be a 1-dimensional list')
    else:
        return f'{identifier.capitalize()} is a 1-dimensional list'

def vector_of_scalars(vector, position = 'only'):
    confirm_vector(vector, position)
    identifier = argument_position(position)
    for scalar in vector:
        if not isinstance(scalar, (int, float)):
            raise TypeError(f'Elements of {identifier} must be integers or floats')
    else:
        return f'{identifier.capitalize()} is a 1-dimensional list containing elements that are integers or floats'

def vector_of_strings(vector, position = 'only'):
    confirm_vector(vector, position)
    identifier = argument_position(position)
    for string in vector:
        if not isinstance(string, str) and string is not None:
            raise TypeError(f'Elements of {identifier} must be strings or None')
    else:
        return f'{identifier.capitalize()} is a 1-dimensional list containing elements that are strings or None'

def compare_vectors(vector_one, vector_two):
    confirm_vector(vector_one, 'first')
    confirm_vector(vector_two, 'second')
    if len(vector_one) is not len(vector_two):
        raise ValueError('Both arguments must contain the same number of elements')
    else:
        return 'Both arguments contain the same number of elements'

def allow_none_vector(vector, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(vector, list):
        raise TypeError(f'{identifier.capitalize()} must be a 1-dimensional list')
    for element in vector:
        if not isinstance(element, (int, float, str)) and element is not None:
            raise TypeError(f'Elements of {identifier} must be integers, floats, strings, or None')
    else:
        return f'Elements of {identifier} are integers, floats, strings, or None'

def length(vector, size):
    if not len(vector) == size:
        raise ValueError(f'Argument must contain exactly {size} elements')
    else:
        return f'Argument contains exactly {size} elements'

def long_vector(vector):
    if not len(vector) >= 10:
        raise ValueError(f'First argument must contain at least 10 elements')
    else:
        return f'First argument contains at least 10 elements'


from regressions.errors.scalars import two_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value

def hyperbolic_derivatives(first_constant, second_constant, precision = 4):
    """
    Calculates the first and second derivatives of a hyperbolic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the reciprocal variable of the original hyperbolic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the constant term of the original hyperbolic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First two arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    derivatives['first']['constants'] : list of float
        Coefficients of the resultant first derivative
    derivatives['first']['evaluation'] : func
        Function for evaluating the resultant first derivative at any float or integer argument; if zero inputted as argument, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    derivatives['second']['constants'] : list of float
        Coefficients of the resultant second derivative
    derivatives['second']['evaluation'] : func
        Function for evaluating the resultant second derivative at any float or integer argument; if zero inputted as argument, it will be converted to a small, non-zero decimal value (e.g., 0.0001)

    See Also
    --------
    :func:`~regressions.analyses.equations.hyperbolic.hyperbolic_equation`, :func:`~regressions.analyses.integrals.hyperbolic.hyperbolic_integral`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.models.hyperbolic.hyperbolic_model`

    Notes
    -----
    - Standard form of a hyperbolic function: :math:`f(x) = a\\cdot{\\frac{1}{x}} + b`
    - First derivative of a hyperbolic function: :math:`f'(x) = -a\\cdot{\\frac{1}{x^2}}`
    - Second derivative of a hyperbolic function: :math:`f''(x) = 2a\\cdot{\\frac{1}{x^3}}`
    - |differentiation_formulas|

    Examples
    --------
    Import `hyperbolic_derivatives` function from `regressions` library
        >>> from regressions.analyses.derivatives.hyperbolic import hyperbolic_derivatives
    Generate the derivatives of a hyperbolic function with coefficients 2 and 3, then display the coefficients of its first and second derivatives
        >>> derivatives_constants = hyperbolic_derivatives(2, 3)
        >>> print(derivatives_constants['first']['constants'])
        [-2.0]
        >>> print(derivatives_constants['second']['constants'])
        [4.0]
    Generate the derivatives of a hyperbolic function with coefficients -2 and 3, then evaluate its first and second derivatives at 10
        >>> derivatives_evaluation = hyperbolic_derivatives(-2, 3)
        >>> print(derivatives_evaluation['first']['evaluation'](10))
        0.02
        >>> print(derivatives_evaluation['second']['evaluation'](10))
        -0.004
    Generate the derivatives of a hyperbolic function with all inputs set to 0, then display the coefficients of its first and second derivatives
        >>> derivatives_zeroes = hyperbolic_derivatives(0, 0)
        >>> print(derivatives_zeroes['first']['constants'])
        [-0.0001]
        >>> print(derivatives_zeroes['second']['constants'])
        [0.0002]
    """
    # Handle input errors
    two_scalars(first_constant, second_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant], precision)

    # Create first derivative
    first_constants = [-1 * coefficients[0]]
    def first_derivative(variable):
        # Circumvent division by zero
        if variable == 0:
            variable = 10**(-precision)
        evaluation = first_constants[0] / variable**2
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    first_dictionary = {
        'constants': first_constants,
        'evaluation': first_derivative
    }

    # Create second derivative
    second_constants = [-2 * first_constants[0]]
    def second_derivative(variable):
        # Circumvent division by zero
        if variable == 0:
            variable = 10**(-precision)
        evaluation = second_constants[0] / variable**3
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    second_dictionary = {
        'constants': second_constants,
        'evaluation': second_derivative
    }

    # Package both derivatives in single dictionary
    results = {
        'first': first_dictionary,
        'second': second_dictionary
    }
    return results


from regressions.errors.scalars import allow_none_scalar, positive_integer
from regressions.errors.vectors import allow_none_vector

def rounded_value(number, precision = 4):
    """
    Rounds a number to a certain decimal place, but returns a non-zero result if the number being rounded is non-zero even if it would round to zero at that level of decimal precision; allows None as a possible input

    Parameters
    ----------
    number : int or float
        Number to round
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    TypeError
        First argument must be an integer, a float, or None
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    number : float
        Original number rounded to the number of decimal places indicated by the precision input

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.summation.sum_value`

    Notes
    -----
    - Number to round: :math:`n`
    - Absolute value of number: :math:`a = |n|`
    - Maximum number of digits after decimal place of result: :math:`d`
    - Check size of number: :math:`c(a,d) = \\lfloor a\\cdot{10^d} \\rfloor`
    - Significant digit: :math:`s(a,d) = \\lfloor ( a\\cdot{10^d} - \\lfloor a\\cdot{10^d} \\rfloor )\\cdot{10} \\rfloor`
    - If :math:`c(a,d) = 0`:
        
        - Rounding formula (if :math:`n = 0`): :math:`r = 0`
        - If :math:`n \\neq 0`:

            - Rounding formula (if :math:`a = n`): :math:`r(d) = 10^{-d}`
            - Rounding formula (if :math:`a \\neq n`): :math:`r(d) = -10^{-d}`
    
    - If :math:`c(a,d) \\neq 0`:

        - If :math:`a = n`:
        
            - Rounding formula (if :math:`s(a,d) \\geq 5`): :math:`r(a,d) = \\frac{\\lceil a\\cdot{10^d} \\rceil}{10^d}`
            - Rounding formula (if :math:`s(a,d) < 5`): :math:`r(a,d) = \\frac{\\lfloor a\\cdot{10^d} \\rfloor}{10^d}`
        
        - If :math:`a \\neq n`:
        
            - Rounding formula (if :math:`s(a,d) \\geq 5`): :math:`r(a,d) = -\\frac{\\lceil a\\cdot{10^d} \\rceil}{10^d}`
            - Rounding formula (if :math:`s(a,d) < 5`): :math:`r(a,d) = -\\frac{\\lfloor a\\cdot{10^d} \\rfloor}{10^d}`
    
    - |rounding|

    Examples
    --------
    Import `rounded_value` function from `regressions` library
        >>> from regressions.statistics.rounding import rounded_value
    Round the number 9.2157823956916472 to six decimal places
        >>> number_normal = rounded_value(9.2157825956916472, 6)
        >>> print(number_normal)
        9.215783
    Round the number -0.00000003 to six decimal places
        >>> number_abnormal = rounded_value(-0.00000003, 6)
        >>> print(number_abnormal)
        -1e-6
    Round the number 11.725371548561 to four decimal places (without providing a value for the precision argument)
        >>> round_skip = rounded_value(11.725371548561)
        >>> print(round_skip)
        11.7254
    """
    # Handle input errors
    allow_none_scalar(number)
    positive_integer(precision)

    # Handle None value
    if number == None:
        return None
    
    # Circumvent rounding to zero with small positive numbers
    elif number < 10**(-precision) and number > 0:
        return 10**(-precision)

    # Circumvent rounding to zero with small negative numbers
    elif number > -10**(-precision) and number < 0:
        return -10**(-precision)
    
    # Handle general case
    else:
        return float(round(number, precision))

def rounded_list(numbers, precision = 4):
    # Handle input errors
    allow_none_vector(numbers, 'first')
    positive_integer(precision)

    # Create list to return
    results = []

    # Iterate over input
    for number in numbers:
        # Store rounded values of input in list to return
        results.append(rounded_value(number, precision))
    
    # Return results
    return results


from regressions.errors.scalars import three_scalars, four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.sort import sorted_list
from regressions.statistics.rounding import rounded_value, rounded_list
from regressions.analyses.derivatives.quadratic import quadratic_derivatives
from .linear import linear_roots

def quadratic_roots(first_constant, second_constant, third_constant, precision = 4):
    """
    Calculates the roots of a quadratic function

    Parameters
    ----------
    first_constant : int or float
        Coefficient of the quadratic term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Coefficient of the linear term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Coefficient of the constant term of the original quadratic function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First three arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    roots : list of float
        List of the x-coordinates of all of the x-intercepts of the original function; if the function never crosses the x-axis, then it will return a list of `None`

    See Also
    --------
    :func:`~regressions.analyses.equations.quadratic.quadratic_equation`, :func:`~regressions.analyses.derivatives.quadratic.quadratic_derivatives`, :func:`~regressions.analyses.integrals.quadratic.quadratic_integral`, :func:`~regressions.models.quadratic.quadratic_model`

    Notes
    -----
    - Standard form of a quadratic function: :math:`f(x) = a\\cdot{x^2} + b\\cdot{x} + c`
    - Quadratic formula: :math:`x = \\frac{-b \\pm \\sqrt{b^2 - 4ac}}{2a}`
    - |quadratic_formula|

    Examples
    --------
    Import `quadratic_roots` function from `regressions` library
        >>> from regressions.analyses.roots.quadratic import quadratic_roots
    Calculate the roots of a quadratic function with coefficients 2, 7, and 5
        >>> roots_first = quadratic_roots(2, 7, 5)
        >>> print(roots_first)
        [-2.5, -1.0]
    Calculate the roots of a quadratic function with coefficients 2, -5, and 3
        >>> roots_second = quadratic_roots(2, -5, 3)
        >>> print(roots_second)
        [1.0, 1.5]
    Calculate the roots of a quadratic function with all inputs set to 0
        >>> roots_zeroes = quadratic_roots(0, 0, 0)
        >>> print(roots_zeroes)
        [None]
    """
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant], precision)

    # Create intermediary list and list to return
    roots = []
    result = []

    # Create intermediary variable
    discriminant = coefficients[1]**2 - 4 * coefficients[0] * coefficients[2]

    # Create roots
    first_root = (-1 * coefficients[1] + discriminant**(1/2)) / (2 * coefficients[0])
    second_root = (-1 * coefficients[1] - discriminant**(1/2)) / (2 * coefficients[0])

    # Eliminate duplicate roots
    if first_root == second_root:
        roots.append(first_root)
    
    # Eliminate complex roots
    else:
        if not isinstance(first_root, complex):
            roots.append(first_root)
        if not isinstance(second_root, complex):
            roots.append(second_root)
    
    # Handle no roots
    if not roots:
        roots.append(None)
    
    # Sort roots
    sorted_roots = sorted_list(roots)
    
    # Round roots
    rounded_roots = rounded_list(sorted_roots, precision)

    # Return result
    result.extend(rounded_roots)
    return result

def quadratic_roots_first_derivative(first_constant, second_constant, third_constant, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)

    # Generate coefficients of first derivative
    constants = quadratic_derivatives(first_constant, second_constant, third_constant)['first']['constants']

    # Determine roots of first derivative
    result = linear_roots(*constants, precision)
    return result

def quadratic_roots_second_derivative(first_constant, second_constant, third_constant, precision = 4):
    # Handle input errors
    three_scalars(first_constant, second_constant, third_constant)
    positive_integer(precision)
    
    # Create list to return
    result = []
    
    # Determine root
    root = None

    # Return result
    result.append(root)
    return result

def quadratic_roots_initial_value(first_constant, second_constant, third_constant, initial_value, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, initial_value)
    positive_integer(precision)

    # Determine roots given an initial value
    result = quadratic_roots(first_constant, second_constant, third_constant - initial_value, precision)
    return result

def quadratic_roots_derivative_initial_value(first_constant, second_constant, third_constant, initial_value, precision = 4):
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, initial_value)
    positive_integer(precision)

    # Determine roots of derivative given an initial value
    result = linear_roots(2 * first_constant, second_constant - initial_value, precision)
    return result


from regressions.errors.matrices import matrix_of_scalars
from regressions.errors.vectors import long_vector
from regressions.errors.scalars import positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.vectors.dimension import single_dimension
from regressions.vectors.column import column_conversion
from regressions.matrices.solve import system_solution
from regressions.analyses.equations.hyperbolic import hyperbolic_equation
from regressions.analyses.derivatives.hyperbolic import hyperbolic_derivatives
from regressions.analyses.integrals.hyperbolic import hyperbolic_integral
from regressions.analyses.points import key_coordinates
from regressions.analyses.accumulation import accumulated_area
from regressions.analyses.mean_values import average_values
from regressions.statistics.summary import five_number_summary
from regressions.statistics.correlation import correlation_coefficient

def hyperbolic_model(data, precision = 4):
    """
    Generates a hyperbolic regression model from a given data set

    Parameters
    ----------
    data : list of lists of int or float
        List of lists of numbers representing a collection of coordinate pairs; it must include at least 10 pairs
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        First argument must contain at least 10 elements
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    model['constants'] : list of float
        Coefficients of the resultant hyperbolic model; the first element is the coefficient of the reciprocal variable term, and the second element is the coefficient of the constant term
    model['evaluations']['equation'] : func
        Function that evaluates the equation of the hyperbolic model at a given numeric input (e.g., model['evaluations']['equation'](10) would evaluate the equation of the hyperbolic model when the independent variable is 10)
    model['evaluations']['derivative'] : func
        Function that evaluates the first derivative of the hyperbolic model at a given numeric input (e.g., model['evaluations']['derivative'](10) would evaluate the first derivative of the hyperbolic model when the independent variable is 10)
    model['evaluations']['integral'] : func
        Function that evaluates the integral of the hyperbolic model at a given numeric input (e.g., model['evaluations']['integral'](10) would evaluate the integral of the hyperbolic model when the independent variable is 10)
    model['points']['roots'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the x-intercepts of the hyperbolic model (will contain exactly one point)
    model['points']['maxima'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the maxima of the hyperbolic model (will always be `None`)
    model['points']['minima'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the minima of the hyperbolic model (will always be `None`)
    model['points']['inflections'] : list of lists of float
        List of lists of numbers representing the coordinate pairs of all the inflection points of the hyperbolic model (will always be `None`)
    model['accumulations']['range'] : float
        Total area under the curve represented by the hyperbolic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided (i.e., over the range)
    model['accumulations']['iqr'] : float
        Total area under the curve represented by the hyperbolic model between the first and third quartiles of all the independent coordinates originally provided (i.e., over the interquartile range)
    model['averages']['range']['average_value_derivative'] : float
        Average rate of change of the curve represented by the hyperbolic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided
    model['averages']['range']['mean_values_derivative'] : list of float
        All points between the smallest independent coordinate originally provided and the largest independent coordinate originally provided where their instantaneous rate of change equals the function's average rate of change over that interval
    model['averages']['range']['average_value_integral'] : float
        Average value of the curve represented by the hyperbolic model between the smallest independent coordinate originally provided and the largest independent coordinate originally provided
    model['averages']['range']['mean_values_integral'] : list of float
        All points between the smallest independent coordinate originally provided and the largest independent coordinate originally provided where their value equals the function's average value over that interval
    model['averages']['iqr']['average_value_derivative'] : float
        Average rate of change of the curve represented by the hyperbolic model between the first and third quartiles of all the independent coordinates originally provided
    model['averages']['iqr']['mean_values_derivative'] : list of float
        All points between the first and third quartiles of all the independent coordinates originally provided where their instantaneous rate of change equals the function's average rate of change over that interval
    model['averages']['iqr']['average_value_integral'] : float
        Average value of the curve represented by the hyperbolic model between the first and third quartiles of all the independent coordinates originally provided
    model['averages']['iqr']['mean_values_integral'] : list of float
        All points between the first and third quartiles of all the independent coordinates originally provided where their value equals the function's average value over that interval
    model['correlation'] : float
        Correlation coefficient indicating how well the model fits the original data set (values range between 0.0, implying no fit, and 1.0, implying a perfect fit)

    See Also
    --------
    :func:`~regressions.analyses.equations.hyperbolic.hyperbolic_equation`, :func:`~regressions.analyses.derivatives.hyperbolic.hyperbolic_derivatives`, :func:`~regressions.analyses.integrals.hyperbolic.hyperbolic_integral`, :func:`~regressions.analyses.roots.hyperbolic.hyperbolic_roots`, :func:`~regressions.statistics.correlation.correlation_coefficient`, :func:`~regressions.execute.run_all`

    Notes
    -----
    - Provided ordered pairs for the data set: :math:`p_i = \\{ (p_{1,x}, p_{1,y}), (p_{2,x}, p_{2,y}), \\cdots, (p_{n,x}, p_{n,y}) \\}`
    - Provided values for the independent variable: :math:`X_i = \\{ p_{1,x}, p_{2,x}, \\cdots, p_{n,x} \\}`
    - Provided values for the dependent variable: :math:`Y_i = \\{ p_{1,y}, p_{2,y}, \\cdots, p_{n,y} \\}`
    - Minimum value of the provided values for the independent variable: :math:`X_{min} \\leq p_{j,x}, \\forall p_{j,x} \\in X_i`
    - Maximum value of the provided values for the independent variable: :math:`X_{max} \\geq p_{j,x}, \\forall p_{j,x} \\in X_i`
    - First quartile of the provided values for the independent variable: :math:`X_{Q1}`
    - Third quartile of the provided values for the independent variable: :math:`X_{Q3}`
    - Mean of all provided values for the dependent variable: :math:`\\bar{y} = \\frac{1}{n}\\cdot{\\sum\\limits_{i=1}^n Y_i}`
    - Resultant values for the coefficients of the hyperbolic model: :math:`C_i = \\{ a, b \\}`
    - Standard form for the equation of the hyperbolic model: :math:`f(x) = a\\cdot{\\frac{1}{x}} + b`
    - First derivative of the hyperbolic model: :math:`f'(x) = -a\\cdot{\\frac{1}{x^2}}`
    - Second derivative of the hyperbolic model: :math:`f''(x) = 2a\\cdot{\\frac{1}{x^3}}`
    - Integral of the hyperbolic model: :math:`F(x) = a\\cdot{\\ln|x|} + b\\cdot{x}`
    - Potential x-values of the roots of the hyperbolic model: :math:`x_{intercepts} = \\{ -\\frac{a}{b} \\}`
    - Potential x-values of the maxima of the hyperbolic model: :math:`x_{maxima} = \\{ 0 \\}`
    - Potential x-values of the minima of the hyperbolic model: :math:`x_{minima} = \\{ 0 \\}`
    - Potential x-values of the inflection points of the hyperbolic model: :math:`x_{inflections} = \\{ 0 \\}`
    - Accumulatation of the hyperbolic model over its range: :math:`A_{range} = \\int_{X_{min}}^{X_{max}} f(x) \\,dx`
    - Accumulatation of the hyperbolic model over its interquartile range: :math:`A_{iqr} = \\int_{X_{Q1}}^{X_{Q3}} f(x) \\,dx`
    - Average rate of change of the hyperbolic model over its range: :math:`m_{range} = \\frac{f(X_{max}) - f(X_{min})}{X_{max} - X_{min}}`
    - Potential x-values at which the hyperbolic model's instantaneous rate of change equals its average rate of change over its range: :math:`x_{m,range} = \\{ -\\sqrt{-\\frac{a}{m_{range}}}, \\sqrt{-\\frac{a}{m_{range}}} \\}`
    - Average value of the hyperbolic model over its range: :math:`v_{range} = \\frac{1}{X_{max} - X_{min}}\\cdot{A_{range}}`
    - Potential x-values at which the hyperbolic model's value equals its average value over its range: :math:`x_{v,range} = \\{ -\\frac{a}{b - v_{range}} \\}`
    - Average rate of change of the hyperbolic model over its interquartile range: :math:`m_{iqr} = \\frac{f(X_{Q3}) - f(X_{Q1})}{X_{Q3} - X_{Q1}}`
    - Potential x-values at which the hyperbolic model's instantaneous rate of change equals its average rate of change over its interquartile range: :math:`x_{m,iqr} = \\{ -\\sqrt{-\\frac{a}{m_{iqr}}}, \\sqrt{-\\frac{a}{m_{iqr}}} \\}`
    - Average value of the hyperbolic model over its interquartile range: :math:`v_{iqr} = \\frac{1}{X_{Q3} - X_{Q1}}\\cdot{A_{iqr}}`
    - Potential x-values at which the hyperbolic model's value equals its average value over its interquartile range: :math:`x_{v,iqr} = \\{ -\\frac{a}{b - v_{iqr}} \\}`
    - Predicted values based on the hyperbolic model: :math:`\\hat{y}_i = \\{ \\hat{y}_1, \\hat{y}_2, \\cdots, \\hat{y}_n \\}`
    - Residuals of the dependent variable: :math:`e_i = \\{ p_{1,y} - \\hat{y}_1, p_{2,y} - \\hat{y}_2, \\cdots, p_{n,y} - \\hat{y}_n \\}`
    - Deviations of the dependent variable: :math:`d_i = \\{ p_{1,y} - \\bar{y}, p_{2,y} - \\bar{y}, \\cdots, p_{n,y} - \\bar{y} \\}`
    - Sum of squares of residuals: :math:`SS_{res} = \\sum\\limits_{i=1}^n e_i^2`
    - Sum of squares of deviations: :math:`SS_{dev} = \\sum\\limits_{i=1}^n d_i^2`
    - Correlation coefficient for the hyperbolic model: :math:`r = \\sqrt{1 - \\frac{SS_{res}}{SS_{dev}}}`
    - |regression_analysis|

    Examples
    --------
    Import `hyperbolic_model` function from `regressions` library
        >>> from regressions.models.hyperbolic import hyperbolic_model
    Generate a hyperbolic regression model for the data set [[1, 2519], [2, 1259], [3, 839], [4, 629], [5, 503], [6, 419], [7, 359], [8, 314], [9, 279], [10, 251]], then print its coefficients, roots, total accumulation over its interquartile range, and correlation
        >>> model_perfect = hyperbolic_model([[1, 2519], [2, 1259], [3, 839], [4, 629], [5, 503], [6, 419], [7, 359], [8, 314], [9, 279], [10, 251]])
        >>> print(model_perfect['constants'])
        [2520.0, -1.0]
        >>> print(model_perfect['points']['roots'])
        [[2520.0, 0.0]]
        >>> print(model_perfect['accumulations']['iqr'])
        2466.6897
        >>> print(model_perfect['correlation'])
        1.0
    Generate a hyperbolic regression model for the data set [[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]], then print its coefficients, inflections, total accumulation over its range, and correlation
        >>> model_agnostic = hyperbolic_model([[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]])
        >>> print(model_agnostic['constants'])
        [-13.5246, 37.7613]
        >>> print(model_agnostic['points']['inflections'])
        [None]
        >>> print(model_agnostic['accumulations']['range'])
        308.7102
        >>> print(model_agnostic['correlation'])
        0.3479
    """
    # Handle input errors
    matrix_of_scalars(data, 'first')
    long_vector(data)
    positive_integer(precision)

    # Store independent and dependent variable values separately
    independent_variable = single_dimension(data, 1)
    dependent_variable = single_dimension(data, 2)

    # Create matrices for independent and dependent variables
    independent_matrix = []
    dependent_matrix = column_conversion(dependent_variable)

    # Iterate over inputted data
    for element in independent_variable:
        # Circumvent division by zero
        if element == 0:
            # Store arbitrarily large number and constant together as lists within independent matrix
            independent_matrix.append([10**precision, 1])
        
        # Handle general case
        else:
            # Store reciprocal and constant evaluations of original independent elements together as lists within independent matrix
            independent_matrix.append([1 / element, 1])
    
    # Solve system of equations
    solution = system_solution(independent_matrix, dependent_matrix, precision)

    # Eliminate zeroes from solution
    coefficients = no_zeroes(solution, precision)

    # Generate evaluations for function, derivatives, and integral
    equation = hyperbolic_equation(*coefficients, precision)
    derivative = hyperbolic_derivatives(*coefficients, precision)['first']['evaluation']
    integral = hyperbolic_integral(*coefficients, precision)['evaluation']

    # Determine key points of graph
    points = key_coordinates('hyperbolic', coefficients, precision)

    # Generate values for lower and upper bounds
    five_numbers = five_number_summary(independent_variable, precision)
    min_value = five_numbers['minimum']
    max_value = five_numbers['maximum']
    q1 = five_numbers['q1']
    q3 = five_numbers['q3']

    # Calculate accumulations
    accumulated_range = accumulated_area('hyperbolic', coefficients, min_value, max_value, precision)
    accumulated_iqr = accumulated_area('hyperbolic', coefficients, q1, q3, precision)

    # Determine average values and their points
    averages_range = average_values('hyperbolic', coefficients, min_value, max_value, precision)
    averages_iqr = average_values('hyperbolic', coefficients, q1, q3, precision)

    # Create list of predicted outputs
    predicted = []
    for element in independent_variable:
        predicted.append(equation(element))
    
    # Calculate correlation coefficient for model
    accuracy = correlation_coefficient(dependent_variable, predicted, precision)
    
    # Package preceding results in multiple dictionaries
    evaluations = {
        'equation': equation,
        'derivative': derivative,
        'integral': integral
    }
    points = {
        'roots': points['roots'],
        'maxima': points['maxima'],
        'minima': points['minima'],
        'inflections': points['inflections']
    }
    accumulations = {
        'range': accumulated_range,
        'iqr': accumulated_iqr
    }
    averages = {
        'range': averages_range,
        'iqr': averages_iqr
    }

    # Package all dictionaries in single dictionary to return
    result = {
        'constants': coefficients,
        'evaluations': evaluations,
        'points': points,
        'accumulations': accumulations,
        'averages': averages,
        'correlation': accuracy
    }
    return result


from regressions.errors.matrices import matrix_of_scalars

def matrix_of_cofactors(matrix):
    """
    Create the matrix of cofactors corresponding to a given matrix

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        Argument must be a 2-dimensional list
    TypeError
        Elements nested within argument must be integers or floats
    
    Returns
    -------
    matrix : list of lists of int or float
        List of lists in which each inner element alternates being positive or negative versions of the corresponding element from the original matrix

    See Also
    --------
    :func:`~regressions.matrices.minors.matrix_of_minors`, :func:`~regressions.matrices.transpose.transposed_matrix`, :func:`~regressions.matrices.determinant.linear_determinant`, :func:`~regressions.matrices.inverse.inverse_matrix`

    Notes
    -----
    - Original matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Matrix of cofactors (if :math:`\\mathbf{A}` contains an odd number of rows and columns): :math:`\\mathbf{A}^C = \\begin{bmatrix} a_{1,1} & -1\\cdot{a_{1,2}} & \\cdots & a_{1,n} \\\\ -1\\cdot{a_{2,1}} & a_{2,2} & \\cdots & -1\\cdot{a_{2,n}} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & -1\\cdot{a_{m,2}} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Matrix of cofactors (if :math:`\\mathbf{A}` contains an even number of rows and columns): :math:`\\mathbf{A}^C = \\begin{bmatrix} a_{1,1} & -1\\cdot{a_{1,2}} & \\cdots & -1\\cdot{a_{1,n}} \\\\ -1\\cdot{a_{2,1}} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ -1\\cdot{a_{m,1}} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - |cofactors|

    Examples
    --------
    Import `matrix_of_cofactors` function from `regressions` library
        >>> from regressions.matrices.cofactors import matrix_of_cofactors
    Create the matrix of cofactors for [[1, 2, 3], [4, 5, 6]]
        >>> matrix_3x2 = matrix_of_cofactors([[1, 2, 3], [4, 5, 6]])
        >>> print(matrix_3x2)
        [[1, -2, 3], [-4, 5, -6]]
    Create the matrix of cofactors for [[2, 3], [5, 7]]
        >>> matrix_2x2 = matrix_of_cofactors([[2, 3], [5, 7]])
        >>> print(matrix_2x2)
        [[2, -3], [-5, 7]]
    """
    # Handle input errors
    matrix_of_scalars(matrix)

    # Create list to return
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix)):
        # Create new lists inside list to return
        result.append([])

        if m % 2 == 0:
            # Iterate over inner lists of input
            for n in range(len(matrix[0])):
                # Handle even-even indexed elements
                if n % 2 == 0:
                    result[m].append(matrix[m][n])
                
                # Handle even-odd indexed elements
                else:
                    result[m].append(-1 * matrix[m][n])
        
        else:
            # Iterate over inner lists of input
            for n in range(len(matrix[0])):
                # Handle odd-even indexed elements
                if n % 2 == 0:
                    result[m].append(-1 * matrix[m][n])
                
                # Handle odd-odd indexed elements
                else:
                    result[m].append(matrix[m][n])
    
    # Return result
    return result


from math import cos
from regressions.errors.scalars import four_scalars, positive_integer
from regressions.errors.adjustments import no_zeroes
from regressions.statistics.rounding import rounded_value, rounded_list

def sinusoidal_integral(first_constant, second_constant, third_constant, fourth_constant, precision = 4):
    """
    Generates the integral of a sinusoidal function

    Parameters
    ----------
    first_constant : int or float
        Vertical stretch factor of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    second_constant : int or float
        Horizontal stretch factor of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    third_constant : int or float
        Horizontal shift of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    fourth_constant : int or float
        Vertical shift of the original sine function; if zero, it will be converted to a small, non-zero decimal value (e.g., 0.0001)
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the resultant roots

    Raises
    ------
    TypeError
        First four arguments must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    integral['constants'] : list of float
        Coefficients of the resultant integral
    integral['evaluation'] : func
        Function for evaluating the resultant integral at any float or integer argument

    See Also
    --------
    :func:`~regressions.analyses.equations.sinusoidal.sinusoidal_equation`, :func:`~regressions.analyses.derivatives.sinusoidal.sinusoidal_derivatives`, :func:`~regressions.analyses.roots.sinusoidal.sinusoidal_roots`, :func:`~regressions.models.sinusoidal.sinusoidal_model`

    Notes
    -----
    - Standard form of a sinusoidal function: :math:`f(x) = a\\cdot{\\sin(b\\cdot(x - c))} + d`
    - Integral of a sinusoidal function: :math:`F(x) = -\\frac{a}{b}\\cdot{\\cos(b\\cdot(x - c))} + d\\cdot{x}`
    - |indefinite_integral|
    - |integration_formulas|
    - |substitution_rule|

    Examples
    --------
    Import `sinusoidal_integral` function from `regressions` library
        >>> from regressions.analyses.integrals.sinusoidal import sinusoidal_integral
    Generate the integral of a sinusoidal function with coefficients 2, 3, 5, and 7, then display its coefficients
        >>> integral_constants = sinusoidal_integral(2, 3, 5, 7)
        >>> print(integral_constants['constants'])
        [-0.6667, 3.0, 5.0, 7.0]
    Generate the integral of a sinusoidal function with coefficients 7, -5, -3, and 2, then evaluate its integral at 10
        >>> integral_evaluation = sinusoidal_integral(7, -5, -3, 2)
        >>> print(integral_evaluation['evaluation'](10))
        19.2126
    Generate the integral of a sinusoidal function with all inputs set to 0, then display its coefficients
        >>> integral_zeroes = sinusoidal_integral(0, 0, 0, 0)
        >>> print(integral_zeroes['constants'])
        [-1.0, 0.0001, 0.0001, 0.0001]
    """
    # Handle input errors
    four_scalars(first_constant, second_constant, third_constant, fourth_constant)
    positive_integer(precision)
    coefficients = no_zeroes([first_constant, second_constant, third_constant, fourth_constant], precision)
    
    # Create constants
    integral_coefficients = [-1 * coefficients[0] / coefficients[1], coefficients[1], coefficients[2], coefficients[3]]
    constants = rounded_list(integral_coefficients, precision)

    # Create evaluation
    def sinusoidal_evaluation(variable):
        evaluation = constants[0] * cos(constants[1] * (variable - constants[2])) + constants[3] * variable
        rounded_evaluation = rounded_value(evaluation, precision)
        return rounded_evaluation
    
    # Package constants and evaluation in single dictionary
    results = {
        'constants': constants,
        'evaluation': sinusoidal_evaluation
    }
    return results


from regressions.errors.matrices import compare_matrices

def matrix_sum(matrix_one, matrix_two):
    """
    Calculates the sum of two matrices

    Parameters
    ----------
    matrix_one : list of lists of int or float
        List of lists of numbers representing a matrix
    matrix_two : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        Arguments must be 2-dimensional lists
    TypeError
        Elements nested within arguments must be integers or floats
    ValueError
        Both arguments must contain the same amount of lists
    ValueError
        First list within first argument and first list within second argument must contain the same amount of elements

    Returns
    -------
    matrix : list of lists of int or float
        List of lists in which each inner element is the sum of the corresponding elements from the input matrices
    
    See Also
    --------
    :func:`~regressions.vectors.addition.vector_sum`, :func:`~regressions.matrices.multiplication.scalar_product_matrix`

    Notes
    -----
    - First matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{bmatrix}`
    - Second matrix: :math:`\\mathbf{B} = \\begin{bmatrix} b_{1,1} & b_{1,2} & \\cdots & b_{1,n} \\\\ b_{2,1} & b_{2,2} & \\cdots & b_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ b_{m,1} & b_{m,2} & \\cdots & b_{m,n} \\end{bmatrix}`
    - Sum of matrices: :math:`\\mathbf{A} + \\mathbf{B} = \\begin{bmatrix} a_{1,1} + b_{1,1} & a_{1,2} + b_{1,2} & \\cdots & a_{1,n} + b_{1,n} \\\\ a_{2,1} + b_{2,1} & a_{2,2} + b_{2,2} & \\cdots & a_{2,n} + b_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} + b_{m,1} & a_{m,2} + b_{m,2} & \\cdots & a_{m,n} + b_{m,n} \\end{bmatrix}`
    - |matrix_addition|

    Examples
    --------
    Import `matrix_sum` function from `regressions` library
        >>> from regressions.matrices.addition import matrix_sum
    Add [[1, 2, 3], [4, 5, 6]] and [[2, 3, 5], [7, 11, 13]]
        >>> matrix_2x3 = matrix_sum([[1, 2, 3], [4, 5, 6]], [[2, 3, 5], [7, 11, 13]])
        >>> print(matrix_2x3)
        [[3, 5, 8], [11, 16, 19]]
    Add [[-2, 5], [7, -1]] and [[8, 2], [-3, 4]]
        >>> matrix_2x2 = matrix_sum([[-2, 5], [7, -1]], [[8, 2], [-3, 4]])
        >>> print(matrix_2x2)
        [[6, 7], [4, 3]]
    """
    # Handle input errors
    compare_matrices(matrix_one, matrix_two)

    # Create list to return
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix_one)):
        # Create new lists inside list to return
        result.append([])

        # Iterate over inner lists of input
        for n in range(len(matrix_one[0])):
            # Store sums in inner lists of return
            result[m].append(matrix_one[m][n] + matrix_two[m][n])
    
    # Return result
    return result


from regressions.errors.vectors import compare_vectors

def vector_sum(vector_one, vector_two):
    """
    Calculates the sum of two vectors

    Parameters
    ----------
    vector_one : list of int or float
        List of numbers representing a vector
    vector_two : list of int or float
        List of numbers representing a vector

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats
    ValueError
        Both arguments must contain the same number of elements

    Returns
    -------
    vector : list of int or float
        List in which each element is the sum of the corresponding elements from the input vectors

    See Also
    --------
    :func:`~regressions.matrices.addition.matrix_sum`, :func:`~regressions.vectors.multiplication.scalar_product_vector`

    Notes
    -----
    - First vector: :math:`\\mathbf{a} = \\langle a_1, a_2, \\cdots, a_n \\rangle`
    - Second vector: :math:`\\mathbf{b} = \\langle b_1, b_2, \\cdots, b_n \\rangle`
    - Sum of vectors: :math:`\\mathbf{a} + \\mathbf{b} = \\langle a_1 + b_1, a_2 + b_2, \\cdots, a_n + b_n \\rangle`
    - |vector_addition|

    Examples
    --------
    Import `vector_sum` function from `regressions` library
        >>> from regressions.vectors.addition import vector_sum
    Add [1, 2, 3] and [4, 5, 6]
        >>> vector_3d = vector_sum([1, 2, 3], [4, 5, 6])
        >>> print(vector_3d)
        [5, 7, 9]
    Add [-5, 12] and [3, -7]
        >>> vector_2d = vector_sum([-5, 12], [3, -7])
        >>> print(vector_2d)
        [-2, 5]
    """
    # Handle input errors
    compare_vectors(vector_one, vector_two)

    # Create list to return
    result = []

    # Iterate over first input
    for i in range(len(vector_one)):
        # Store sums of corresponding vector elements in result
        result.append(vector_one[i] + vector_two[i])
    
    # Return result
    return result


from regressions.errors.vectors import vector_of_scalars
from regressions.errors.scalars import positive_integer
from .minimum import minimum_value
from .maximum import maximum_value
from .median import median_value
from .quartiles import quartile_value
from .rounding import rounded_value

def five_number_summary(data, precision = 4):
    """
    Calculates the five number summary of a given data set: minimum, first quartile, median, third quartile, and maximum

    Parameters
    ----------
    data : list of int or float
        List of numbers to analyze
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the result

    Raises
    ------
    TypeError
        First argument must be a 1-dimensional list
    TypeError
        Elements of first argument must be integers or floats
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    summary['minimum'] : float
        Smallest value from the data set
    summary['q1'] : float
        First quartile of the data set, below which 25% of the data fall
    summary['median'] : float
        Middle value of the data set, splitting the data evenly in half
    summary['q3'] : float
        Third quartile of the data set, above which 25% of the data fall
    summary['maximum'] : float
        Largest value from the data set

    See Also
    --------
    :func:`~regressions.statistics.sort.sorted_list`, :func:`~regressions.statistics.minimum.minimum_value`, :func:`~regressions.statistics.maximum.maximum_value`, :func:`~regressions.statistics.median.median_value`, :func:`~regressions.statistics.quartiles.quartile_value`

    Notes
    -----
    - Set of numbers: :math:`a_i = \\{ a_1, a_2, \\cdots, a_n \\}`
    - Minimum: :math:`a_{min} \\leq a_j, \\forall a_j \\in a_i`
    - Maximum: :math:`a_{max} \\geq a_j, \\forall a_j \\in a_i`
    - For sets with an odd amount of numbers:
        
        - First quartile: :math:`Q_1 = a_{\\lceil n/4 \\rceil}`
        - Median: :math:`M = a_{\\lceil n/2 \\rceil}`
        - Third quartile: :math:`Q_3 = a_{\\lceil 3n/4 \\rceil}`
    
    - For sets with an even amount of numbers:

        - If :math:`n \\text{ mod } 4 \\neq 0`:

            - First quartile: :math:`Q_1 = a_{\\lceil n/4 \\rceil}`
            - Median: :math:`M = \\frac{a_{n/2} + a_{n/2 + 1}}{2}`
            - Third quartile: :math:`Q_3 = a_{\\lceil 3n/4 \\rceil}`
        
        - If :math:`n \\text{ mod } 4 = 0`:

            - First quartile: :math:`Q_1 = \\frac{a_{n/4} + a_{n/4 + 1}}{2}`
            - Median: :math:`M = \\frac{a_{n/2} + a_{n/2 + 1}}{2}`
            - Third quartile: :math:`Q_3 = \\frac{a_{3n/4} + a_{3n/4 + 1}}{2}`

    - |five_number_summary|

    Examples
    --------
    Import `five_number_summary` function from `regressions` library
        >>> from regressions.statistics.summary import five_number_summary
    Determine the five number summary of the set [21, 53, 3, 68, 43, 9, 72, 19, 20, 1]
        >>> summary_even = five_number_summary([21, 53, 3, 68, 43, 9, 72, 19, 20, 1])
        >>> print(summary_even['q1'])
        9.0
        >>> print(summary_even['maximum'])
        72.0
    Determine the five number summary of the set [12, 81, 13, 8, 42, 72, 91, 20, 20]
        >>> summary_odd = five_number_summary([12, 81, 13, 8, 42, 72, 91, 20, 20])
        >>> print(summary_odd['q3'])
        76.5
        >>> print(summary_odd['minimum'])
        8.0
    """
    # Handle input errors
    vector_of_scalars(data, 'first')
    positive_integer(precision)

    # Calculate all values used in five-number summary
    min_value = minimum_value(data)
    q1 = quartile_value(data, 1)
    median = median_value(data)
    q3 = quartile_value(data, 3)
    max_value = maximum_value(data)

    # Round all values
    rounded_min = rounded_value(min_value, precision)
    rounded_q1 = rounded_value(q1, precision)
    rounded_med = rounded_value(median, precision)
    rounded_q3 = rounded_value(q3, precision)
    rounded_max = rounded_value(max_value, precision)
    
    # Package values in single dictionary to return
    result = {
        'minimum': rounded_min,
        'q1': rounded_q1,
        'median': rounded_med,
        'q3': rounded_q3,
        'maximum': rounded_max
    }
    return result


from regressions.errors.matrices import square_matrix
from .determinant import linear_determinant, inner_determinant

def matrix_of_minors(matrix):
    """
    Create the matrix of minors corresponding to a given matrix

    Parameters
    ----------
    matrix : list of lists of int or float
        List of lists of numbers representing a matrix

    Raises
    ------
    TypeError
        Argument must be a 2-dimensional list
    TypeError
        Elements nested within argument must be integers or floats
    ValueError
        Argument must contain the same amount of lists as the amount of elements contained within its first list
    
    Returns
    -------
    matrix : list of lists of float
        List of lists in which each inner element is a determinant of a subsection of the original matrix

    See Also
    --------
    :func:`~regressions.matrices.cofactors.matrix_of_cofactors`, :func:`~regressions.matrices.transpose.transposed_matrix`, :func:`~regressions.matrices.determinant.linear_determinant`, :func:`~regressions.matrices.inverse.inverse_matrix`

    Notes
    -----
    - Original matrix: :math:`\\mathbf{A} = \\begin{bmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,j} & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,j} & a_{2,n} \\\\ a_{i,1} & a_{i,2} & \\cdots & a_{i,j} & a_{i,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,j} & a_{m,n} \\end{bmatrix}`
    - Minor of matrix: :math:`|\\mathbf{A}_{i,j}| = \\begin{vmatrix} a_{1,1} & a_{1,2} & \\cdots & a_{1,n} \\\\ a_{2,1} & a_{2,2} & \\cdots & a_{2,n} \\\\ \\cdots & \\cdots & \\cdots & \\cdots \\\\ a_{m,1} & a_{m,2} & \\cdots & a_{m,n} \\end{vmatrix}`
    - Matrix of minors: :math:`\\mathbf{A}^M = \\begin{bmatrix} |\\mathbf{A}_{1,1}| & |\\mathbf{A}_{1,2}| & \\cdots & |\\mathbf{A}_{1,j}| & |\\mathbf{A}_{1,n}| \\\\ |\\mathbf{A}_{2,1}| & |\\mathbf{A}_{2,2}| & \\cdots & |\\mathbf{A}_{2,j}| & |\\mathbf{A}_{2,n}| \\\\ |\\mathbf{A}_{i,1}| & |\\mathbf{A}_{i,2}| & \\cdots & |\\mathbf{A}_{i,j}| & |\\mathbf{A}_{i,n}| \\\\ \\cdots & \\cdots & \\cdots & \\cdots & \\cdots \\\\ |\\mathbf{A}_{m,1}| & |\\mathbf{A}_{m,2}| & \\cdots & |\\mathbf{A}_{m,j}| & |\\mathbf{A}_{m,n}| \\end{bmatrix}`
    - |minors|

    Examples
    --------
    Import `matrix_of_minors` function from `regressions` library
        >>> from regressions.matrices.minors import matrix_of_minors
    Create the matrix of minors for [[1, 2], [3, 4]]
        >>> matrix_2x2 = matrix_of_minors([[1, 2], [3, 4]])
        >>> print(matrix_2x2)
        [[4.0, 3.0], [2.0, 1.0]]
    Create the matrix of minors for [[2, 3, 5], [7, 11, 13], [17, 19, 23]]
        >>> matrix_3x3 = matrix_of_minors([[2, 3, 5], [7, 11, 13], [17, 19, 23]])
        >>> print(matrix_3x3)
        [[6.0, -60.0, -54.0], [-26.0, -39.0, -13.0], [-16.0, -9.0, 1.0]]
    """
    # Handle input errors
    square_matrix(matrix)

    # Create list to return
    result = []

    # Iterate over outer lists of input
    for m in range(len(matrix)):
        # Create new lists inside list to return
        result.append([])

        # Iterate over inner lists of input
        for n in range(len(matrix[0])):
            # Store determinants in inner lists of return
            result[m].append(linear_determinant(inner_determinant(matrix, m, n)))
    
    # Return result
    return result


from .positions import argument_position

def scalar_value(scalar, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(scalar, (int, float)):
        raise TypeError(f'{identifier.capitalize()} must be an integer or a float')
    else:
        return f'{identifier.capitalize()} is an integer or a float'

def two_scalars(scalar_one, scalar_two):
    scalar_value(scalar_one, 'first')
    scalar_value(scalar_two, 'second')
    return 'Both first and second arguments are integers or floats'

def three_scalars(scalar_one, scalar_two, scalar_three):
    two_scalars(scalar_one, scalar_two)
    scalar_value(scalar_three, 'third')
    return 'First, second, and third arguments are all integers or floats'

def four_scalars(scalar_one, scalar_two, scalar_three, scalar_four):
    three_scalars(scalar_one, scalar_two, scalar_three)
    scalar_value(scalar_four, 'fourth')
    return 'First, second, third, and fourth arguments are all integers or floats'

def five_scalars(scalar_one, scalar_two, scalar_three, scalar_four, scalar_five):
    four_scalars(scalar_one, scalar_two, scalar_three, scalar_four)
    scalar_value(scalar_five, 'fifth')
    return 'First, second, third, fourth, and fifth arguments are all integers or floats'

def compare_scalars(scalar_one, scalar_two, position_one, position_two):
    scalar_value(scalar_one, position_one)
    scalar_value(scalar_two, position_two)
    if scalar_one > scalar_two:
        raise ValueError(f'{position_one.capitalize()} argument must be less than or equal to {position_two} argument')
    else:
        return f'{position_one.capitalize()} argument is less than or equal to {position_two} argument'

def positive_integer(scalar):
    if not isinstance(scalar, int) or not scalar > 0:
        raise ValueError('Last argument must be a positive integer')
    else:
        return 'Last argument is a positive integer'

def whole_number(scalar, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(scalar, int) or not scalar >= 0:
        raise ValueError(f'{identifier.capitalize()} must be a whole number')
    else:
        return f'{identifier.capitalize()} is a whole number'

def select_integers(scalar, choices, position = 'only'):
    identifier = argument_position(position)
    if scalar not in choices:
        raise ValueError(f'{identifier.capitalize()} must be one of the following integers: {choices}')
    else:
        return f'{identifier.capitalize()} is one of the following integers: {choices}'

def allow_none_scalar(scalar, position = 'only'):
    identifier = argument_position(position)
    if not isinstance(scalar, (int, float)) and scalar is not None:
        raise TypeError(f'{identifier.capitalize()} must be an integer, a float, or None')
    else:
        return f'{identifier.capitalize()} is an integer, a float, or None'


from .errors.matrices import matrix_of_scalars
from .errors.vectors import long_vector
from .errors.scalars import positive_integer
from .models.linear import linear_model
from .models.quadratic import quadratic_model
from .models.cubic import cubic_model
from .models.hyperbolic import hyperbolic_model
from .models.exponential import exponential_model
from .models.logarithmic import logarithmic_model
from .models.logistic import logistic_model
from .models.sinusoidal import sinusoidal_model
from .vectors.dimension import single_dimension
from .statistics.minimum import minimum_value
from .statistics.maximum import maximum_value
from .statistics.quartiles import quartile_value
from .statistics.mean import mean_value
from .statistics.median import median_value

def run_all(data, precision = 4):
    """
    Generates all eight key regression models (linear, quadratic, cubic, hyperbolic, exponential, logarithmic, logistic, and sinusoidal) for a given data set, in addition to determining the best fit based on correlation and providing various statistical measures

    Parameters
    ----------
    data : list of lists of int or float
        List of lists of numbers representing a collection of coordinate pairs; it must include at least 10 pairs
    precision : int, default=4
        Maximum number of digits that can appear after the decimal place of the results

    Raises
    ------
    TypeError
        First argument must be a 2-dimensional list
    TypeError
        Elements nested within first argument must be integers or floats
    ValueError
        First argument must contain at least 10 elements
    ValueError
        Last argument must be a positive integer

    Returns
    -------
    results['models']['linear'] : dict
        See :ref:`Linear Model`
    results['models']['quadratic'] : dict
        See :ref:`Quadratic Model`
    results['models']['cubic'] : dict
        See :ref:`Cubic Model`
    results['models']['hyperbolic'] : dict
        See :ref:`Hyperbolic Model`
    results['models']['exponential'] : dict
        See :ref:`Exponential Model`
    results['models']['logarithmic'] : dict
        See :ref:`Logarithmic Model`
    results['models']['logistic'] : dict
        See :ref:`Logistic Model`
    results['models']['sinusoidal'] : dict
        See :ref:`Sinusoidal Model`
    results['statistics']['minimum'] : int or float
        Smallest value of the independent variable from the provided data set
    results['statistics']['maximum'] : int or float
        Largest value of the independent variable from the provided data set
    results['statistics']['q1'] : int or float
        First quartile of the independent variable from the provided data set, below which 25% of the data points lie
    results['statistics']['q3'] : int or float
        Third quartile of the independent variable from the provided data set, above which 25% of the data points lie
    results['statistics']['mean'] : int or float
        Arithmetic mean of the independent variable from the provided data set
    results['statistics']['median'] : int or float
        Median of the independent variable from the provided data set, which splits the data in half
    results['optimal']['option'] : str
        Name of the model with the highest correlation for this particular data set (e.g., 'cubic' if the cubic model has a higher correlation than all other seven models)
    results['optimal']['correlation'] : float
        Value of the correlation for the model with the best fit (i.e., the model listed in ['optimal']['option'])

    See Also
    --------
    :func:`~regressions.models.linear.linear_model`, :func:`~regressions.models.quadratic.quadratic_model`, :func:`~regressions.models.cubic.cubic_model`, :func:`~regressions.models.hyperbolic.hyperbolic_model`, :func:`~regressions.models.exponential.exponential_model`, :func:`~regressions.models.logarithmic.logarithmic_model`, :func:`~regressions.models.logistic.logistic_model`, :func:`~regressions.models.sinusoidal.sinusoidal_model`, :func:`~regressions.statistics.summary.five_number_summary`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Provided ordered pairs for the data set: :math:`p_i = \\{ (p_{1,x}, p_{1,y}), (p_{2,x}, p_{2,y}), \\cdots, (p_{n,x}, p_{n,y}) \\}`
    - Provided values for the independent variable: :math:`X_i = \\{ p_{1,x}, p_{2,x}, \\cdots, p_{n,x} \\}`
    - Provided values for the dependent variable: :math:`Y_i = \\{ p_{1,y}, p_{2,y}, \\cdots, p_{n,y} \\}`
    - Resultant values for the coefficients of the linear model: :math:`C_{lin} = \\{ a_{lin}, b_{lin} \\}`
    - Standard form for the equation of the linear model: :math:`lin(x) = a_{lin}\\cdot{x} + b_{lin}`
    - Resultant values for the coefficients of the quadratic model: :math:`C_{quad} = \\{ a_{quad}, b_{quad}, c_{quad} \\}`
    - Standard form for the equation of the quadratic model: :math:`quad(x) = a_{quad}\\cdot{x^2} + b_{quad}\\cdot{x} + c_{quad}`
    - Resultant values for the coefficients of the cubic model: :math:`C_{cub} = \\{ a_{cub}, b_{cub}, c_{cub}, d_{cub} \\}`
    - Standard form for the equation of the cubic model: :math:`cub(x) = a_{cub}\\cdot{x^3} + b_{cub}\\cdot{x^2} + c_{cub}\\cdot{x} + d_{cub}`
    - Resultant values for the coefficients of the hyperbolic model: :math:`C_{hyp} = \\{ a_{hyp}, b_{hyp} \\}`
    - Standard form for the equation of the hyperbolic model: :math:`hyp(x) = a_{hyp}\\cdot{\\frac{1}{x}} + b_{hyp}`
    - Resultant values for the coefficients of the exponential model: :math:`C_{exp} = \\{ a_{exp}, b_{exp} \\}`
    - Standard form for the equation of the exponential model: :math:`exp(x) = a_{exp}\\cdot{b_{exp}^x}`
    - Resultant values for the coefficients of the logarithmic model: :math:`C_{log} = \\{ a_{log}, b_{log} \\}`
    - Standard form for the equation of the logarithmic model: :math:`log(x) = a_{log}\\cdot{\\ln{x}} + b_{log}`
    - Resultant values for the coefficients of the logistic model: :math:`C_{lst} = \\{ a_{lst}, b_{lst}, c_{lst} \\}`
    - Standard form for the equation of the logistic model: :math:`lst(x) = \\frac{a_{lst}}{1 + \\text{e}^{-b_{lst}\\cdot(x - c_{lst})}}`
    - Resultant values for the coefficients of the sinusoidal model: :math:`C_{sin} = \\{ a_{sin}, b_{sin}, c_{sin}, d_{sin} \\}`
    - Standard form for the equation of the sinusoidal model: :math:`sin(x) = a_{sin}\\cdot{\\sin(b_{sin}\\cdot(x - c_{sin}))} + d_{sin}`
    - |regression_analysis|

    Examples
    --------
    Import `run_all` function from `regressions` library
        >>> from regressions.execute import run_all
    Generate all eight regression models for the data set [[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]], then print each model's coefficients, the mean of the data set, and the name of the model with the best fit
        >>> results = run_all([[1, 32], [2, 25], [3, 14], [4, 23], [5, 39], [6, 45], [7, 42], [8, 49], [9, 36], [10, 33]])
        >>> print(results['models']['linear']['constants'])
        [1.9636, 23.0]
        >>> print(results['models']['quadratic']['constants'])
        [-0.3106, 5.3803, 16.1667]
        >>> print(results['models']['cubic']['constants'])
        [-0.3881, 6.0932, -24.155, 49.4667]
        >>> print(results['models']['hyperbolic']['constants'])
        [-13.5246, 37.7613]
        >>> print(results['models']['exponential']['constants'])
        [22.1049, 1.0692]
        >>> print(results['models']['logarithmic']['constants'])
        [7.4791, 22.5032]
        >>> print(results['models']['logistic']['constants'])
        [43.983, 0.3076, 0.9746]
        >>> print(results['models']['sinusoidal']['constants'])
        [14.0875, 0.7119, -3.7531, 34.2915]
        >>> print(results['statistics']['mean'])
        5.5
        >>> print(results['optimal']['option'])
        'sinusoidal'
    Generate all eight regression models for the data set [[169, 423], [122, 391], [178, 555], [131, 284], [120, 520], [179, 558], [164, 265], [167, 338], [198, 445], [139, 402], [183, 725], [133, 470], [156, 573], [159, 325], [121, 653], [118, 358], [122, 633], [167, 487], [161, 453], [194, 488], [170, 517], [124, 377], [191, 310], [194, 398], [173, 744], [166, 389], [113, 583], [109, 380], [126, 668], [144, 491], [107, 533], [188, 355], [147, 553], [169, 497], [121, 606], [132, 373], [111, 554], [173, 669], [177, 483], [122, 340], [171, 286], [108, 681], [139, 502], [115, 339], [174, 396], [134, 625], [147, 435], [146, 555], [147, 656], [126, 354], [155, 679], [181, 629], [149, 417], [119, 374], [102, 422], [112, 292], [108, 464], [109, 559], [112, 635], [159, 518], [180, 304], [185, 567], [165, 299], [160, 337], [133, 730], [193, 374], [164, 537], [172, 592], [173, 660], [186, 290], [170, 670], [192, 687], [154, 596], [154, 464], [125, 383], [193, 559], [155, 586], [149, 406], [131, 590], [127, 339], [163, 378], [145, 254], [156, 395], [166, 355], [189, 661], [133, 685], [168, 685], [190, 736], [145, 564], [125, 470], [129, 541], [133, 439], [162, 486], [125, 387], [183, 596], [135, 733], [106, 329], [100, 279], [102, 439], [162, 454]], then print each model's coefficients, the mean of the data set, and the name of the model with the best fit
        >>> results_large = run_all([[169, 423], [122, 391], [178, 555], [131, 284], [120, 520], [179, 558], [164, 265], [167, 338], [198, 445], [139, 402], [183, 725], [133, 470], [156, 573], [159, 325], [121, 653], [118, 358], [122, 633], [167, 487], [161, 453], [194, 488], [170, 517], [124, 377], [191, 310], [194, 398], [173, 744], [166, 389], [113, 583], [109, 380], [126, 668], [144, 491], [107, 533], [188, 355], [147, 553], [169, 497], [121, 606], [132, 373], [111, 554], [173, 669], [177, 483], [122, 340], [171, 286], [108, 681], [139, 502], [115, 339], [174, 396], [134, 625], [147, 435], [146, 555], [147, 656], [126, 354], [155, 679], [181, 629], [149, 417], [119, 374], [102, 422], [112, 292], [108, 464], [109, 559], [112, 635], [159, 518], [180, 304], [185, 567], [165, 299], [160, 337], [133, 730], [193, 374], [164, 537], [172, 592], [173, 660], [186, 290], [170, 670], [192, 687], [154, 596], [154, 464], [125, 383], [193, 559], [155, 586], [149, 406], [131, 590], [127, 339], [163, 378], [145, 254], [156, 395], [166, 355], [189, 661], [133, 685], [168, 685], [190, 736], [145, 564], [125, 470], [129, 541], [133, 439], [162, 486], [125, 387], [183, 596], [135, 733], [106, 329], [100, 279], [102, 439], [162, 454]])
        >>> print(results_large['models']['linear']['constants'])
        [0.4934, 414.5401]
        >>> print(results_large['models']['quadratic']['constants'])
        [-0.007, 2.5668, 265.4919]
        >>> print(results_large['models']['cubic']['constants'])
        [0.0005, -0.2204, 33.8099, -1226.1398]
        >>> print(results_large['models']['hyperbolic']['constants'])
        [-10786.2465, 563.019]
        >>> print(results_large['models']['exponential']['constants'])
        [407.8094, 1.0009]
        >>> print(results_large['models']['logarithmic']['constants'])
        [74.0076, 118.997]
        >>> print(results_large['models']['logistic']['constants'])
        [488.2, 1.0, 1.0]
        >>> print(results_large['models']['sinusoidal']['constants'])
        [32.3199, 1.0085, 1.8848, 488.9635]
        >>> print(results_large['statistics']['mean'])
        149.29
        >>> print(results_large['optimal']['option'])
        'sinusoidal'
    """
    # Handle input errors
    matrix_of_scalars(data, 'first')
    long_vector(data)
    positive_integer(precision)

    # Grab values of independent variable
    independent_variable = single_dimension(data, 1)

    # Generate all eight key regression models
    models = {
        'linear': linear_model(data, precision),
        'quadratic': quadratic_model(data, precision),
        'cubic': cubic_model(data, precision),
        'hyperbolic': hyperbolic_model(data, precision),
        'exponential': exponential_model(data, precision),
        'logarithmic': logarithmic_model(data, precision),
        'logistic': logistic_model(data, precision),
        'sinusoidal': sinusoidal_model(data, precision)
    }

    # Determine key statistical values for independent variable
    statistics = {
        'minimum': minimum_value(independent_variable),
        'maximum': maximum_value(independent_variable),
        'q1': quartile_value(independent_variable, 1),
        'q3': quartile_value(independent_variable, 3),
        'mean': mean_value(independent_variable),
        'median': median_value(independent_variable)
    }

    # Grab correlations of all previously generated models
    correlations = {
        'linear': models['linear']['correlation'],
        'quadratic': models['quadratic']['correlation'],
        'cubic': models['cubic']['correlation'],
        'hyperbolic': models['hyperbolic']['correlation'],
        'exponential': models['exponential']['correlation'],
        'logarithmic': models['logarithmic']['correlation'],
        'logistic': models['logistic']['correlation'],
        'sinusoidal': models['sinusoidal']['correlation']
    }

    # Determine model with highest correlation
    best = max(correlations, key=correlations.get)
    optimal = {
        'option': best,
        'correlation': correlations[best]
    }

    # Package preceding results in single dictionary to return
    result = {
        'models': models,
        'statistics': statistics,
        'optimal': optimal
    }
    return result


from regressions.errors.scalars import scalar_value
from regressions.errors.vectors import vector_of_scalars
from .mean import mean_value

def single_deviation(actual, mean): 
    """
    Calculates the difference between the actual value from a data set and the mean of that data set

    Parameters
    ----------
    actual : int or float
        Value actually provided by an initial data set
    mean : int or float
        Average value of the data set

    Raises
    ------
    TypeError
        Arguments must be integers or floats

    Returns
    -------
    deviation : float
        Difference between the actual value and the mean of the data set

    See Also
    --------
    :func:`~regressions.statistics.mean.mean_value`, :func:`~regressions.statistics.residuals.single_residual`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Observed value: :math:`y`
    - Mean of all observed values: :math:`\\bar{y}`
    - Deviation: :math:`d = y - \\bar{y}`
    - |deviation|

    Examples
    --------
    Import `single_deviation` function from `regressions` library
        >>> from regressions.statistics.deviations import single_deviation
    Determine the deviation for an actual value of 7.8 and a mean of 13.75
        >>> deviation_small = single_deviation(7.8, 13.75)
        >>> print(deviation_small)
        -5.95
    Determine the deviation between an actual value of 6.1 and an mean of -19.7
        >>> deviation_large = single_deviation(6.1, -19.7)
        >>> print(deviation_large)
        25.799999999999997
    """
    # Handle input errors
    scalar_value(actual, 'first')
    scalar_value(mean, 'second')

    # Calcuate difference between actual and mean
    difference = actual - mean

    # Convert difference to float
    result = float(difference)
    return result

def multiple_deviations(actual_array):
    """
    Generates a list of the differences between the actual values from an original list and mean value of the actual values from that original list

    Parameters
    ----------
    actual_array : list of int or float
        List containing the actual values observed from a data set

    Raises
    ------
    TypeError
        Arguments must be 1-dimensional lists
    TypeError
        Elements of arguments must be integers or floats

    Returns
    -------
    deviations : list of float
        List of differences between the actual values and the mean value for all elements from the original list

    See Also
    --------
    :func:`~regressions.statistics.mean.mean_value`, :func:`~regressions.statistics.residuals.multiple_residuals`, :func:`~regressions.statistics.correlation.correlation_coefficient`

    Notes
    -----
    - Observed values: :math:`y_i = \\{ y_1, y_2, \\cdots, y_n \\}`
    - Mean of all observed values: :math:`\\bar{y} = \\frac{1}{n}\\cdot{\\sum\\limits_{i=1}^n y_i}`
    - Deviations: :math:`d_i = \\{ y_1 - \\bar{y}, y_2 - \\bar{y}, \\cdots, y_n - \\bar{y} \\}`
    - |deviation|

    Examples
    --------
    Import `multiple_deviations` function from `regressions` library
        >>> from regressions.statistics.deviations import multiple_deviations
    Generate a list of deviations from this data set [8.2, 9.41, 1.23, 34.7]
        >>> deviations_short = multiple_deviations([8.2, 9.41, 1.23, 34.7])
        >>> print(deviations_short)
        [-5.185000000000002, -3.9750000000000014, -12.155000000000001, 21.315]
    Generate a list of deviations from this data set [5.21, 8.2, 9.41, 1.23, 10.52, 21.76, 34.7]
        >>> deviations_long = multiple_deviations([5.21, 8.2, 9.41, 1.23, 10.52, 21.76, 34.7])
        >>> print(deviations_long)
        [-7.7942857142857145, -4.804285714285715, -3.5942857142857143, -11.774285714285714, -2.484285714285715, 8.755714285714287, 21.69571428571429]
    """
    # Handle input errors
    vector_of_scalars(actual_array)

    # Create list to return
    results = []

    # Calculate mean of input
    average = mean_value(actual_array)

    # Iterate over input
    for element in actual_array:
        # Store deviation of each element in list to return
        results.append(single_deviation(element, average))
    
    # Return results
    return results